/*********************************************************************
 * File : Biconn.c
 *
 * Biconnected parts of vertices.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"

/* Entry points */
Public  bool    biconn          ARG((NoArg));

/* External references */
/* ergrf.c */
Extern  Void    erenter    ARG((string));
Extern  Void    erreturn   ARG((string));

/* local functions */
Private Void    biinit   ARG((NoArg));
Private Void    bisearch ARG((node,node,int*,int*,int*,int*,int*));
Private node    binext   ARG((node,node,node));
Private bool    biskip   ARG((int,int,int,int,int,int,bool));

/* local variables */
Local   int     bicount;     /* count the number of node */
Local   int     bivdef[MAXNODES];
Local   int     bivlow[MAXNODES];
Local   bool    biexit;

#define bidef(v)   bivdef[v]
#define bilow(v)   bivlow[v]
#define BINEW     -1

/*==================================================================*/
Public  bool    biconn(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int nd, jl, je, jb, jk, jc;

#if DBG
    erenter("biconn");
#endif

    if(loop < 1 || !expand)
        return True;

    if((nd = binext(0, -1, -1)) == NDNULL) {
        return True;
    }
    biinit();
    bisearch(nd, -1, &jl, &je, &jb, &jk, &jc);

#if DBG
    if(dbg >= bidbg) {
        printf("*dbg* biconn: graph %ld, loop = %d(%d), ",
               nconfig, loop, jl);
        printf("extern = %d(%d)", nextern, je);

        printf(", iline = %d bridge = %d, block = %d, type=%d\n",
               loop+nnodes-1, jb, jk, biexts*2 + bitadp);
    }
    erreturn("biconn");
#endif

    if(biexit)
        return False;

    /* No check of the answer in the test version */
    return True;
}


/*==================================================================*/
Private Void    biinit(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v;

    bicount = 0;
    biexit  = False;
    bitadp  = False;
    bi1pi   = True;
    biblock = True;

    biexts  = False;
    for(v = 0; v < nnodes; v++) {
        bidef(v) = BINEW;
        bilow(v) = BINEW;
    }
}

/*==================================================================*/
Private Void    bisearch(v, ov, iloop, iext, ibrg, iblk, icbrg)
node    v, ov;
int     *iloop, *iext, *ibrg, *iblk, *icbrg;
/*--------------------------------------------------------------------
 *   Search graph to fined bridge.
 *
 *   Arguments:
 *       v : input  : node to visit.
 *       ov : input  : the last node.
 *       iloop : output : no. of loop below.
 *       iext  : output : no. of external particles below.
 *       ibrg  : output : no. of bridges below.
 *       icbrg : output : no. of bridges of the current opi component.
 */
{
    node w;
    int il, ie, ib, ik, ic, lp;

#if DBG
    if(dbg >= bidbg)
        printf("*dbg* bisearch : begin v = %d, ov = %d\n", v, ov);
#endif

    /* visit node "v" */
    bicount++;
    *iext     = 0;
    *ibrg     = 0;
    bidef(v)  = bicount;
    bilow(v)  = bicount;
    *iloop    = ndloop(v);
    *iblk     = 0;
    *icbrg    = 0;

    /* visit sons */
    for(w = binext(v, ov, -1); w != NDNULL; w = binext(v, ov, w)) {
        ic = 0;
        if(w == v) {    /* self loop */
            (*iloop) += edgetbl[v][w] / 2;
            (*iblk)  += edgetbl[v][w] / 2;

        } else if(ndtype(w) == NDEXTERN) {    /* external particle */
            (*iext)++;
            ic = 1;

        } else if(bidef(w) != BINEW) {  /* edge (v, w) is a back edge */
            bilow(v)  = min(bilow(v), bidef(w));
            if(bidef(v) < bidef(w))
                (*iloop) += edgetbl[v][w];

        } else {    /* new vertex */
            bisearch(w, v, &il, &ie, &ib, &ik, &ic);
            if(biexit)
                return;
            (*iloop) += il;
            (*iext)  += ie;
            (*ibrg)  += ib;
            (*iblk)  += ik;

            lp = ndloop(w);

            if(bilow(w) >= bidef(v)) { /* cut point */
                (*iblk)++;
                if(block && biskip(*ibrg, *iblk, ie, il,
                                      -1, lp, False)) {
                    biexit = True;
                    return;
                }
            }
            if(bilow(w) > bidef(v)) {
                if(edgetbl[v][w] == 1) { /* bridge */
                    (*ibrg)++;
                    ic++;
                    if(biskip(*ibrg, *iblk, ie, il,
                              ic, lp, True)) {
                        biexit = True;
                        return;
                    }
                    ic = 1;
                } else {                 /* multiple connection */
                    (*iloop) += edgetbl[v][w] - 1;
                }
            } else {                     /* not bridge */
                bilow(v) = min(bilow(v), bilow(w));
                (*iloop) += edgetbl[v][w] - 1;
            }
        }
        *icbrg += ic;
#if DBG
        if(dbg >= bidbg) {
            printf("*dbg* bisearch : %d->(%d, %d) = %d",
                   ov, v, w, edgetbl[v]);
            printf(", il=%d,ie=%d,ib=%d,ik=%d\n",
                   *iloop,*iext,*ibrg,*iblk);
        }
#endif
    } /* of for */
    if(ndfree(v) > 0)
        bilow(v) = -1;
}

/*==================================================================*/
Private node    binext(v, ov, lv)
node    v, ov, lv;
/*--------------------------------------------------------------------
 *  Find sons of "v", where father of "v" is "ov".
 *  The latest found son is "lv".
 */
{
    node  w;

    for(w = lv + 1; w < nnodes; w++) {
        if(w == ov)
            continue;
        if(edgetbl[v][w] > 0)
            return w;
    }
    return NDNULL;
}

/*==================================================================*/
Private bool    biskip(brg, blk, ext, lp, cb, wlp, brgf)
int     brg, blk, ext, lp, cb, wlp;
bool    brgf;
/*--------------------------------------------------------------------
 *  Test if delete the current file.
 *  Selection rules are given by flags.
 */
{
    biexit  = False;
    bitadp  = False;
    bi1pi   = True;
    biblock = True;

    if(blk > 1)
        biblock = False;

    if(brg > 0)
        bi1pi = False;

    if(brgf) {
        if(ext < 1)
            bitadp = True;

        else if(ext < 2)
            biexts = True;

        if(lp <= loop-1 && ext == nextern-1)
            biexts = True;
    }
#if DBG
    if(dbg>=bidbg) {
        printf("biskip: bitadp=%d, biexts=%d, bi1pi=%d, biblock=%d\n",
                bitadp   , biexts   , bi1pi   , biblock  );
        printf("selfe = %d, cb = %d, wlp = %d\n", selfe, cb, wlp);
    }
#endif

    if(!selfe && cb == 2 && wlp == 0) {
#if DBG
        if(dbg>=bidbg)
            printf("biskip: return True by selfe\n");
#endif
        return True;
    }

    if(!tadpole && bitadp) {
#if DBG
        if(dbg>=bidbg)
            printf("biskip: return True by tadpole\n");
#endif
        return True;
    }

    if(!extself && biexts) {
#if DBG
        if(dbg>=bidbg)
            printf("biskip: return True by extself\n");
#endif
        return True;
    }

    if(onepi && !bi1pi) {
#if DBG
        if(dbg>=bidbg)
            printf("biskip: return True by onepi\n");
#endif
        return True;
    }

    if(block && !biblock) {
#if DBG
        if(dbg>=bidbg)
            printf("biskip: return True by block\n");
#endif
        return True;
    }

#if DBG
    if(dbg>=bidbg)
        printf("biskip: return False\n");
#endif

    return False;
}
