/********************************************************************
 *  File : elem.c
 *
 *  Operations on graph elements
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"

/*------------------------------------------------------------------*/
/* Entry points : they should be called after pvint and pvread      */
/* 1. initialization */
Public  Void    elinit          ARG((NoArg));

/* 2. creation of nodes */
Public  Void    eldelv          ARG((int, int, int));
Public  Void    elgene          ARG((pindex, int));
Public  Void    elgenv          ARG((int, int, int));

/* 3. end of node creation */
Public  Void    elgnfin         ARG((NoArg));

/* 4. in the connecting process */
Public  Void    elconn          ARG((node, node, int));
Public  Void    eldisc          ARG((node, node, int));
Public  bool    elleft          ARG((NoArg));
Public  int     elncmp          ARG((node, node));
Public  int     elccmp          ARG((node, node, int, bool));
Public  int     elroot          ARG((NoArg));
Public  long    elmult          ARG((node, node));
Public  bool    elcable         ARG((NoArg));

/* 5. any time */
Public  Void    eldump          ARG((NoArg));
Public  string  elstype         ARG((int));

/* 6. after calling painit */
Public int      eledge          ARG((node, node, pindex*, int*));

/*------------------------------------------------------------------*/
/* External references */
/* ergrf.c */
Extern  Void    erfail     ARG((NoArg));
Extern  Void    error      ARG((string, string));
/* print.c */
Extern  Void    prdump          ARG((NoArg));
/* rdptbl.c */
Extern  char   *pvname          ARG((pindex));

/*------------------------------------------------------------------*/
/* Private functions */
Private int     ellevel         ARG((node));
Private int     elcpcmp         ARG((int,pindex,int,pindex));

/*------------------------------------------------------------------*/
/* Local variable */
Local   bool    elalloc = False;

/*==================================================================*/
Public Void   elinit(NoArg)
/*--------------------------------------------------------------------
 *  Initialize the module of elements.
 *  This function may be called several times.
 */
{
    int i, j, msize, bptr;
    char *ptr;

    passg    = False;
    nnodes   = 0;
    nlgtbl   = 0;
    nconfig  = 0;
    nprocess = 0;


    if(!elalloc) {      /* for the first call */
        elalloc = True;
        msize =   sizeof(tgraph)
                + sizeof(tnode) * MAXNODES
                + sizeof(tleg)  * MAXLEGS  * MAXNODES;

        if((ptr = (char*) malloc(msize)) == NULL) {
            error("elinit", "memory allocation failed(1).");
            erfail();
        }
        cgraph        = (tgraph*) ptr;
        bptr          = sizeof(tgraph);
        cgraph->basen = bptr;
        bptr         += sizeof(tnode)*MAXNODES;
        cgraph->offsl = bptr;


        for(i = 0; i < MAXNODES; i++) {
            if((nodetbl[i] = (tdnode*)malloc(sizeof(tdnode)))
               == NULL) {
                error("elinit", "memory allocation failed(1).");
                erfail();
            }
            if((pntbl[i] = (tpntbl*)malloc(sizeof(tpntbl))) == NULL) {
                error("elinit", "memory allocation failed(2).");
                erfail();
            }

            ndtype(i)  = NDNULL;
        }
    } else {
        for(i = 0; i < MAXNODES; i++) {
            ndtype(i)  = NDNULL;
        }
    }
#if DBG
    for(i = 0; i < MAXNODES; i++) {
        for(j = 0; j < MAXNODES; j++) {
            nddeg(i, j) = 0;
        }
        ndtype(i) = NDNULL;
    }
#endif
    for(i = 0; i < MAXNODES; i++) {
        ndlevel(i) = NDINF;
        for(j = 0; j < MAXNODES; j++) {
            edgetbl[i][j] = 0;
        }
    }
}

/*==================================================================*/
Public Void   elgene(ptcl, infi)
pindex  ptcl;
int     infi;
/*--------------------------------------------------------------------
 *  Create external particle
 *
 *  Argument
 *      ptcl    : input : particle name
 *      infi    : input : "initial" or "final"
 */
{
    node   n;


    /* find empty slot in the node table and store */
    for(n = 0; n < MAXNODES; n++) {
        if(ndtype(n) == NDNULL) {
            nnodes++;


            ndbasel(n) = cgraph->offsl + nlgtbl*sizeof(tleg);
            nlgtbl++;

            ndtype(n)  = NDEXTERN;
            ndlegs(n)  = 1;
            ndloop(n)  = 0;
            ndfree(n)  = 1;
            ndundef(n) = 0;

            ndname(n)  = ptcl;

            /* normalize to FINAL */
            ndiorf(n)  = infi;
            if(infi == ELINITIAL)
                ndname(n) *= ptable[abs(ptcl)]->antip;

            return;
        }
    }

    /* the node table is full */
    error("elgene", "too many nodes");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   elgenv(lg, lp, n)
int lg;
int lp;
int n;
/*--------------------------------------------------------------------
 *  Generate nodes.
 *
 *  Arguments:
 *    lg : input : the number of legs of the nodes.
 *    lp : input : the number of loops of the nodes.
 *    n  : input : the number of nodes to be generated.
 */
{
    node v;
    int p;

    for(v = 0; v < MAXNODES; v++) {
        if(ndtype(v) == NDNULL) {
            nnodes++;

            ndbasel(v)  = cgraph->offsl + nlgtbl*sizeof(tleg);
            nlgtbl     += lg;

            ndtype(v)   = (lp == 0) ? NDTREEV : NDLOOPEDV;
            ndlegs(v)   = lg;
            ndfree(v)   = lg;
            ndloop(v)   = lp;
            ndvertex(v) = VNULL;
            ndundef(v)  = lg;
            for(p = 0; p < lg; p++)
                ndparticle(v, p) = PNULL;

            if(--n < 1)
                return;
        }
    }
    error("elgenv", "too many nodes.");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   eldelv(lg, lp, n)
int lg;
int lp;
int n;
/*--------------------------------------------------------------------
 *  Delete nodes.
 *
 *  Arguments:
 *    lg : input : the number of legs of the node.
 *    lp : input : the number of loops of the node.
 *    n  : input : the number of nodes to be deleted.
 */
{
    node m;
    int  type;

    type = (lp == 0) ? NDTREEV : NDLOOPEDV;
    for(m = MAXNODES-1; m >= 0; m--) {
        if(ndtype(m) == type && ndlegs(m) == lg && ndloop(m) == lp){
            ndtype(m) = NDNULL;
            nnodes--;
            if((--n) < 1)
                return;
        }
    }
    error("eldelv", "inconsistent configuration.");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   elgnfin(NoArg)
/*--------------------------------------------------------------------
 *  End of the node generation
 */
{
    int linner, legs;
    int i, j;

    /* check consistency */
    legs    = 0;
    linner  = 0;
    for (i = 0; i < nnodes; i++) {
        legs    += ndlegs(i);
        linner  += ndloop(i);
    }

    j  = nnodes - legs/2 + loop - linner - 1;
    if(legs%2 != 0 || j != 0) {
        error("elgnfin", "inconsistent configuration:");
        fprintf(Stderr, " i2=%d:even, j=%d:0\n", legs, 0);
        fprintf(Stderr, " nextern = %d, legs = %d, nnodes = %d,",
                        nextern, legs, nnodes);
        fprintf(Stderr, " loop = %d, linner = %d\n", loop, linner);
        eldump();
        erfail();
    }
}


/*==================================================================*/
Public Void   elconn(n1, n2, legs)
node n1, n2;
int  legs;
/*--------------------------------------------------------------------
 * connect two nodes with arg3 legs.
 * Configuration is saved.
 */
{
#if DBG
    if(edgetbl[n1][n2] != 0 || edgetbl[n2][n1] != 0) {
        error("elconn", "inconsistent connection.");
        fprintf(Stderr, "(%d,%d,%d) %d %d\n", n1, n2, legs,
                edgetbl[n1][n2], edgetbl[n2][n1]);
        prdump();
        erfail();
    }
#endif
    if(n1 == n2) {
        ndfree(n1)      -= 2*legs;
        edgetbl[n1][n1]  = 2*legs;
    } else {
        ndfree(n1)      -= legs;
        ndfree(n2)      -= legs;
        edgetbl[n1][n2]  = legs;
        edgetbl[n2][n1]  = legs;
        ndlevel(n1)      = min(ndlevel(n1), ndlevel(n2) + 1);
        ndlevel(n2)      = min(ndlevel(n2), ndlevel(n1) + 1);
    }
}

/*==================================================================*/
Public Void   eldisc(n1, n2, legs)
node n1, n2;
int  legs;
/*--------------------------------------------------------------------
 * disconnect two nodes with arg3 legs.
 * The last configuration is restored.
 */
{
    if(n1 == n2) {
#if DBG
        if(edgetbl[n1][n1] != 2*legs) {
            error("eldisc", "inconsistent connection.");
            fprintf(Stderr, "(%d,%d,%d)\n", n1, n2, legs);
            prdump();
            erfail();
        }
#endif
        edgetbl[n1][n1]  = 0;
        ndfree(n1)      += 2*legs;
    } else if(edgetbl[n1][n2] != legs || edgetbl[n2][n1] != legs) {
#if DBG
        error("eldisc", "inconsistent connection.");
        fprintf(Stderr, "(%d,%d,%d)\n", n1, n2, legs);
        prdump();
        erfail();
#endif
    } else {
        edgetbl[n1][n2]  = 0;
        edgetbl[n2][n1]  = 0;
        ndfree(n1)      += legs;
        ndfree(n2)      += legs;
        ndlevel(n1)      = ellevel(n1);
        ndlevel(n2)      = ellevel(n2);
    }
}

/*==================================================================*/
Public bool   elleft(NoArg)
/*--------------------------------------------------------------------
 * Test whether if nodes are left with free legs.
 * This routine should be changed to used table (switch with #define?)
 */
{
    node n;

    for(n = 0; n < nnodes; n++) {
        if(ndfree(n) > 0) {
            return True;
        }
    }
    return False;
}

/*==================================================================*/
Private int     ellevel(n)
node n;
/*--------------------------------------------------------------------
 * calculate level of the node "n" by the edge table.
 */
{
    int         lvl;
    node        m;

    lvl = NDINF;        /* level - 1 */

    if(n == 0)
        return 0;
    for(m = 0; m < nnodes; m++) {
        if(m != n && edgetbl[n][m] > 0) {
            lvl = min(lvl, ndlevel(m)+1);
        }
    }
    return lvl;
}

/*==================================================================*/
Public int    elncmp(nd1, nd2)
node nd1, nd2;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering
 * of its table elements e for "s" <= e < "n".
 */
{
    int  r;

    if((r = ndtype(nd1) - ndtype(nd2)) != 0)
        return r;

    if((r = ndlegs(nd1) - ndlegs(nd2)) != 0)
        return r;

    if((r = ndloop(nd1) - ndloop(nd2)) != 0)
        return r;

    return 0;

}

/*==================================================================*/
Public int    elccmp(nd1, nd2, n, assign)
node nd1, nd2;
int  n;
bool assign;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering
 * of its table elements : e < "n".
 */
{
    int  v;
    long r;

    for(v = 0; v < n; v++) {
        if((r = nddeg(nd1, v) - nddeg(nd2, v)) != 0) {
            return r;
        }
    }
    if(assign)
        return elpcmp(nd1, nd2);
    else
        return 0;

}

/*==================================================================*/
Private int    elcpcmp(c1, p1, c2, p2)
int    c1;
pindex p1;
int    c2;
pindex p2;
/*--------------------------------------------------------------------
 */
{
    int r;

    if((r = c1 - c2) != 0)
        return r;
    return (p1 - p2);
}

/*==================================================================*/
Public  int    elpcmp(nd1, nd2)
node nd1, nd2;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering.
 * This function compares by assigned particles.
 * This is callable only after completing particle assignmnet.
 */
{
    node   nd[2];
    int     elcl[2][MAXTL];
    pindex  elpt[2][MAXTL];
    vindex n;
    int    l, i, cl, lg, r;
    pindex pt;

    nd[0] = nd1;
    nd[1] = nd2;
    lg = ndlegs(nd1);
    if(lg != ndlegs(nd2))
        return (lg - ndlegs(nd2));

    /* ??? for looped vertex ??? */
    if(lg > MAXTL) {
        fprintf(Stderr,"*** elpcmp : too many legs, lg = %d\n", lg);
        exit(16);
    }
    
    /* construct sorted table of particles and l-class
     * of opposit node for each leg
     */
    for(n = 0; n < 2; n++) {
        for(l = 0; l < lg; l++) {
            cl = ndclass(ndnextn(nd[n], l));
            pt = ndparticle(nd[n], l);

            /* find the position to be inserted */
            for(i = l - 1; i >= 0; i--) {
                if(elcpcmp(cl, pt, elcl[n][i], elpt[n][i]) < 0) {
                    elcl[n][i+1] = elcl[n][i];
                    elpt[n][i+1] = elpt[n][i];
                } else {
                    break;
                }
            }
            elcl[n][i+1] = cl;
            elpt[n][i+1] = pt;
        }
    }

#if DBG
    if(dbg > padbg) {
        for(n = 0; n < 2; n++) {
            printf("elpcmp : nd%d=%d : ", n, nd[n]);
            for(l = 0; l < lg; l++)
                printf("(%d,%d) ", elcl[n][l], elpt[n][l]);
            putchar(EOL);
        }
    }
#endif

    for(l = 0; l < lg; l++) {
        r = elcpcmp(elcl[0][l], elpt[0][l], elcl[1][l], elpt[1][l]);
#if DBG
        if(dbg > padbg)
            printf("elpcmp : r=%d \n",r);
#endif
        if(r != 0) {
            return r;
        }
    }
    return r;
}

/*==================================================================*/
Public int   elroot(NoArg)
/*--------------------------------------------------------------------
 * determines root and returns its level.
 */
{
    if(ndtype(0) != NDEXTERN) {
        error("elroot", "no root is found.");
        prdump();
        erfail();
    }
    ndlevel(0) = 0;
    return 0;
}

/*==================================================================*/
Public  long   elmult(nd1, nd2)
node    nd1, nd2;
/*--------------------------------------------------------------------
 * returns multiplicity of the connection between "nd1" and "nd2".
 * The value is coded as ("MAXTL" + 1)**multiplicity.
 *
 * Check for overflow is necessary.
 */
{
    long        val;
    int         i;

    if(edgetbl[nd1][nd2] < 1)
        return 0;

    val = 1;
    for(i = 0; i < edgetbl[nd1][nd2] - 1; i++) {
        val *= MAXTL+1;
    }
    if(nd1 != nd2 && val > 1)
        val *= MAXTL;        /* assume max(deg(n1)) <= 4 */
    return val;
}

/*==================================================================*/
Public  bool    elcable(NoArg)
/*--------------------------------------------------------------------
 * test if there are free legs in the connected components
 * including the root.
 */
{
    node n;
    bool ans;

    ans = True;
    for(n = 0; n < nnodes; n++) {
        if(ndlevel(n) != NDINF) {
            if(ndfree(n) > 0)
                return True;
        } else {
            ans = False;
        }
    }
    return ans;
}

/*==================================================================*/
Public Void   eldump(NoArg)
/*--------------------------------------------------------------------
 *  Print the current configurations of graphs elements.
 *  Table of nodes are printed in "lcdump".
 */
{
    node        n1, n2;

    printf("--------------------------\n");
    printf(" The number of nodes = %d\n", nnodes);

    printf("---- external line table ----\n");
    for(n1 = 0; n1 < nnodes; n1++)
        if(ndtype(n1) == NDEXTERN) {
            printf("%2d %s %s\n", n1, eliorf(ndiorf(n1)),
                   pvname(ndname(n1)));
        }
    putchar(EOL);

    printf("---- connection table ----\n");
    printf("      ");
    for(n2 = 0; n2 < nnodes; n2++) {
        printf("%4d", n2);
    }
    putchar(EOL);

    for(n1 = 0; n1 < nnodes; n1++) {
        printf("%4d :", n1);
        for(n2 = 0; n2 < nnodes; n2++) {
            printf("%4d", edgetbl[n1][n2]);
        }
        putchar(EOL);
    }
}

/*==================================================================*/
Public  string  elstype(tp)
int     tp;
/*--------------------------------------------------------------------
 */
{
    if(tp == NDEXTERN)
        return "external line";
    else if(tp == NDTREEV)
        return "tree vertex  ";
    else if(tp == NDLOOPEDV)
        return "looped vertex";
    else
        return "undefined    ";
}

/*==================================================================*/
Public int      eledge(nd1, nd2, ptcl, mult)
node   nd1, nd2;
pindex *ptcl;
int    *mult;
/*--------------------------------------------------------------------
 *  make table edges going from nd1 to nd2.
 *  ptcl[] : is the sorted list of particles.
 *  mult[] : is the multiplicity 
 *  the size of ptcl is returned.
 */
{
    int    legs;
    int    m, l, i;
    int    np;
    pindex p;

    if(ndtype(nd1) == NDEXTERN)
        legs = 1;
    else
        legs = ndlegs(nd1);

    np = 0;
    for(l = 0; l < legs; l++) {
        if(ndnextn(nd1, l) == nd2) {
            p = ndparticle(nd1, l);

            i = np;
            for(m = 0; m < np; m++) {
                if(ptcl[m] > p) {
                    i = m;
                    break;
                } else if(ptcl[m] == p) {
                    i = -1;
                    mult[m]++;
                    break;
                }
            }
            /* new particle */
            if(i >= 0) {
                for(m = np-1; m >= i; m--) {
                    ptcl[m+1] = ptcl[m];
                    mult[m+1] = mult[m];
                }
                ptcl[i] = p;
                mult[i] = 1;
                np++;
            }
        }
    }
    return np;
}
