/*********************************************************************
 *  File : flgrf.c
 *
 *  Construct the table of fermion lines.
 *
 *  Only spin=1/2 is checked.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"

/* Entry points */
Public  Void    fldump   ARG((NoArg));
Public  int     flfact   ARG((NoArg));
Public  bool    flfurry  ARG((NoArg));
Public  Void    flget    ARG((tgraph*));
Public  bool    flmajo   ARG((NoArg));
Public  Void    flmdir   ARG((NoArg));

/* External references */
/* ergrf.c */
Extern  Void    erfail     ARG((NoArg));
Extern  Void    error      ARG((string, string));
/* rdptbl.c */
Extern  int     pvboson  ARG((pindex));
Extern  int     pvfermi  ARG((pindex));

/* local functions */
Private Void    flinit   ARG((NoArg));
Private int     flfind   ARG((int*));
Private Void    fldmpt   ARG((int,int));
Private int     flleg    ARG((node));
Private int     flmleg   ARG((node));
Private int     flnextl  ARG((node,int));
Private int     floutlg  ARG((node,node));
Private bool    flundef  ARG((NoArg));

#define flin(n,l,p)  ((*(p)=ndparticle((n),(l))) < 0 \
                     && (!pvboson(-(*(p)))))
#define flout(n,l,p) ((*(p)=ndparticle((n),(l))) > 0 \
                     && (!pvboson((*(p)))))
#define floutd(n,l,p) ((*(p)=ndparticle((n),(l))) > 0 \
                      && (pvfermi((*(p)))))
#define flmaj(p)     (ptable[p]->ptype == P_MAJORANA)
#define flfermi(p)   (ptable[p]->ptype == P_FERMION)

/* local variables */
Local bool floccupied = False;
Local bool majorana;
Local node flnodes[MAXNODES]; /* temporary list of nodes */
Local int  fllegs [MAXNODES]; /* temporary list of legs */

/* the number of incoming fermion legs 
 * the number of legs if majorana particle appears in the graph
 */
Local node flflegs[MAXNODES]; 

#define fltnam(typ)  (((typ)==FLOPEN)?"open  ": \
                      ((typ)==FLCLOSED)?"closed":"undef ")
/*==================================================================*/
Public  Void    flget(grf)
tgraph *grf;
/*--------------------------------------------------------------------
 *    Construct the table of f-lines.
 */
{
    int f, nn, n, ntyp;

    cgraph = grf;

    if(flundef())
        return;

    /* free used area */
    if(floccupied) {
        for(f = nflines-1; f >= 0; f--) {
            if(fltable[f].length > 0) {
                if(fltable[f].ndlist != NULL) {
                    free((AnyP)fltable[f].ndlist);
                    free((AnyP)fltable[f].lglist);
                }
            }
        }
    } else {
        floccupied = True;
    }

    flinit();

    nflines = 0;
    while((nn = flfind(&ntyp)) > 0) {
#if DBG
        if(dbg >= fldbg)
            printf("flget: flfind = %d, ntype= %s\n",nn,fltnam(ntyp));
#endif
        fltable[nflines].ndlist = (node *) malloc(nn*sizeof(node));
        if(fltable[nflines].ndlist == NULL) {
            error("flget","Memory allocation failed.");
            erfail();
        }
        fltable[nflines].lglist = (int *) malloc(nn*sizeof(node));
        if(fltable[nflines].lglist == NULL) {
            error("flget","Memory allocation failed.");
            erfail();
        }
        fltable[nflines].type   = ntyp;
        fltable[nflines].length = nn;
        for(n = 0; n < nn; n++) {
            fltable[nflines].ndlist[n] = flnodes[n];
            fltable[nflines].lglist[n] = fllegs[n];
        }

        /* determine direction of i-lines of majorana */

        nflines++;
    }


#if DBG
    if(dbg >= fldbg)
        fldump();
#endif
}

/*==================================================================*/
Private bool    flundef(NoArg)
/*--------------------------------------------------------------------
 *    Check whether if Majorana particle appears in the graph or not.
 */
{
    node    n;
    int     l, nlg;
    pindex  p;

    for(n = 0; n < ndnodes(); n++) {
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            if(ndparticle(n, l) == 0)
                return True;
        }
    }
    return False;
}

/*==================================================================*/
Public  bool    flmajo(NoArg)
/*--------------------------------------------------------------------
 *    Check whether if Majorana particle appears in the graph or not.
 */
{
    node    n;
    int     l, nlg;
    pindex  p;

    for(n = 0; n < ndnodes(); n++) {
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            p = abs(ndparticle(n, l));
            if(flmaj(p)) {
                return True;
            }
        }
    }
    return False;
}

/*==================================================================*/
Private Void    flinit(NoArg)
/*--------------------------------------------------------------------
 *    Initialization
 */
{
    node    n;
    int     l, nlg;
    pindex  p;

    majorana = flmajo();

#if DBG
    if(dbg >= fldbg)
        printf("majorana = %d\n", majorana);
#endif

    for(n = 0; n < ndnodes(); n++) {
        flflegs[n] = 0;
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            if(majorana) {
                p = abs(ndparticle(n, l));
                if(flfermi(p) || flmaj(p))
                    ndilwk(n,l) = True;
                else
                    ndilwk(n,l) = False;
            } else {
                if(flin(n, l, &p))
                    flflegs[n]++;
            }
        }
    }
}

/*==================================================================*/
Private int     flfind(ntyp)
int    *ntyp;
/*--------------------------------------------------------------------
 *    Finds a fermion line and returns the length.
 */
{
    node    n, nv, start, nn;
    int     nl, l, lv, ll;
    pindex p;

#if DBG
    if(dbg >= fldbg) {
        if(majorana) {
            for(nn = 0; nn < ndnodes(); nn++) {
                printf("%4d : ", nn);
                for(ll = 0; ll < ndlegs(nn); ll++) {
                    printf(" %d:%d", ll, ndilwk(nn, ll));
                }
                printf("\n");
            }
        }
    }
#endif

    *ntyp = FLUNDEF;

    /* find start of f-line.  Search external lines first */
    lv = 0;
    for(n = 0; n < ndnodes(); n++) {
        if(majorana) {
            if(ndtype(n) == NDEXTERN && ndilwk(n, 0)) {
                lv = 0;
                start = nv = n;
                *ntyp = FLOPEN;
                if(floutd(n, 0, &p))
                    break;
            } else if((l = flmleg(n)) >= 0) {
                if(ndtype(n) == NDONSHELL || *ntyp == FLUNDEF) {
                    lv    = l;
                    start = nv = n;
                    *ntyp = FLCLOSED;
                }
            }
        } else {
            if(ndtype(n) == NDEXTERN) {
                if(flflegs[n] == 0 && flout(n, 0, &p)) {
                    flflegs[n] = -1;
                    start = nv = n;
                    *ntyp = FLOPEN;
                    break;
                }
            } else if(flflegs[n] > 0) {
                start = nv = n;
                *ntyp = FLCLOSED;
                if(ndtype(n) == NDONSHELL)
                    break;
            }
        }
    }

#if DBG
    if(dbg >= fldbg) {
        if(majorana)
            printf("flfind:1: start=%d, type=%d, lv=%d\n", start, *ntyp,lv);
        else
            printf("flfind:2: start=%d, type=%d\n", start, *ntyp);
    }
#endif

    /* found : external particles */
    if(*ntyp == FLOPEN || (majorana && *ntyp != FLUNDEF)) {
        flnodes[0] = start;
        fllegs[0]  = lv;
        nl         = 1;
        l          = lv;
        if(majorana) {
            ndilwk(nv, l) = False;
            lv            = flnextl(nv, l);
        }
        nv         = ndnextn(nv, l);
#if DBG
        if(dbg >= fldbg) {
            printf("flfind:3: nv=%d, type=%d, lv=%d\n", nv, *ntyp,lv);
        }
#endif
        
    /* found : vertex */
    } else if(*ntyp != FLUNDEF) {
        nl         = 0;

    /* not found */
    } else {
        return 0;
    }

    /* search nodes */
    for(n = 0; n <= ndnodes(); n++) { /* maximum number of repeat */

#if DBG
        if(dbg >= fldbg) {
            printf("flfind:4: %d: nv = %d\n", nl, nv);
            if(majorana) {
                printf("node = %d, leg = %d\n", nv, lv);
                for(nn = 0; nn < ndnodes(); nn++) {
                    printf("%4d : ", nn);
                    for(ll = 0; ll < ndlegs(nn); ll++) {
                        printf(" %d:%d", ll, ndilwk(nn, ll));
                    }
                    printf("\n");
                }
            }
        }
#endif

        if(majorana) {
            if(ndilwk(nv, lv)) {
                ndilwk(nv, lv) = False;
            } else {
                error("flfind", "cannot enter to the node");
                fprintf(Stderr, "node = %d, leg = %d\n", nv, lv);
                erfail();
            }
        }

        /* check external particle nv */
        if(ndtype(nv) == NDEXTERN) {
    
            /* terminate at an external particle */
            if(majorana && *ntyp == FLOPEN) {
                flnodes[nl] = nv;
                fllegs[nl]  = 0;
                nl++;
                ndilwk(nv, 0) = False;
                return nl;
    
            } else if(flflegs[nv] == 1 && *ntyp == FLOPEN) {
                flnodes[nl] = nv;
                fllegs[nl]  = 0;  /* ??? */
                nl++;
                --flflegs[nv];
                return nl;
    
            /* fermion loop reached to an external particle 
             * or cannot go to the external line nv
             */
            } else {
                error("flfind", "inconsistent fermion number");
                fprintf(Stderr, "external particle = %d\n", nv);
                fldmpt(nl,*ntyp);
                erfail();
            }
    
        /* check loop is closed */
        } else if(nl > 0 && start == nv) {
            if(!majorana) {
                --flflegs[nv];
            }
#if DBG
            if(dbg >= fldbg) {
                printf("flfind:5: %d: nv = %d\n", nl, nv);
                if(majorana) {
                    printf("node = %d, leg = %d\n", nv, lv);
                    for(nn = 0; nn < ndnodes(); nn++) {
                        printf("%4d : ", nn);
                        for(ll = 0; ll < ndlegs(nn); ll++) {
                            printf(" %d:%d", ll, ndilwk(nn, ll));
                        }
                        printf("\n");
                    }
                }
            }
#endif
            return nl;
    
        /* cannot go into nv */
        } else if(majorana) {
            if((l = flmleg(nv)) >= 0) {
                ndilwk(nv, l) = False;
                flnodes[nl] = nv;
                fllegs[nl]  = l;
                nl++;
                lv = flnextl(nv, l);
                nv = ndnextn(nv, l);
#if DBG
               if(dbg >= fldbg)
                  printf("flfind:6: %d: nv = %d\n", nl, nv);
#endif
            } else {
                error("flfind", "inconsistent fermion number.");
                fprintf(Stderr, "vertex = %d\n", nv);
                fldmpt(nl,*ntyp);
                erfail();
            }
    
        } else if(flflegs[nv] < 1) {
            error("flfind", "inconsistent fermion number.");
            fprintf(Stderr, "vertex = %d\n", nv);
            fldmpt(nl,*ntyp);
            erfail();
    
        /* search ndlegs(nv)-th outgoing fermion leg */
        } else {
            flnodes[nl] = nv;
            fllegs[nl]  = l = flleg(nv);
            --flflegs[nv];
            nl++;
#if DBG
           if(dbg >= fldbg)
              printf("flfind:8: %d: nv = %d, l=%d\n", nl, nv,l);
#endif
            nv = ndnextn(nv, l);
#if DBG
           if(dbg >= fldbg)
              printf("flfind:8: %d: nv = %d\n", nl, nv);
#endif
        }
    }
    error("flfind", "inconsistent search");
    fldmpt(nl,*ntyp);
    erfail();
    return -1;
}

/*==================================================================*/
Private  int    flleg(nv)
node nv;
/*--------------------------------------------------------------------
 *   ????
 */
{
    int ni, l;
    pindex  p;

    ni = 0;
    for(l = 0; l < ndlegs(nv); l++) {
        if(flout(nv, l, &p)) {
            ni++;
            if(ni == flflegs[nv])
                return l;
        }
    }
    if(l >= ndlegs(nv)) {
        error("flleg", "no outgoing fermion.");
        fprintf(Stderr, "vertex = %d\n", nv);
        erfail();
    }
    return -1;
}

/*==================================================================*/
Private  int    flmleg(nv)
node nv;
/*--------------------------------------------------------------------
 *   Find Majorana leg 
 */
{
    int l;

    for(l = 0; l < ndlegs(nv); l++) {
        if(ndilwk(nv, l)) {
            return l;
        }
    }
    return -1;
}

/*==================================================================*/
Public  bool    flfurry(NoArg)
/*--------------------------------------------------------------------
 *    test Furry's theorem.
 *    This function should be called after "flget".
 */
{
    int     f, n, l, nf=0;
    node    nd;
    pindex  p;
    int     vi;

    /* looped fermion lines with odd number of lengths */
    for(f = 0; f < nflines; f++) {
        if(fltable[f].type == FLOPEN)
            continue;
        if(fltable[f].length % 2 == 0)
            continue;
        vi = True;
        for(n = 0; n < fltable[n].length; n++) {
            nd = fltable[f].ndlist[n];
            for(l = 0; l < ndlegs(nd); l++) {
                p = abs(ndparticle(nd, l));
                if(ptable[p]->ptype == P_FERMION) {
                    continue;
                }
                if(!ptable[p]->furry) {
                    vi = False;
                    break;
                }
            }
            if(vi) {
#if DBG
                if(dbg >= fldbg)
                    printf("delete by Furry : f=%d\n", f);
#endif
                return False;
            }
        }
    }

    /* looped vertex */
    for(n = 0; n < ndnodes(); n++) {
        if(ndtype(n) != NDLOOPEDV)
            continue;
        if(ndlegs(n) % 2 == 0)
            continue;

        vi = True;
        nf = 0;
        for(l = 0; l < ndlegs(n); l++) {
            p = abs(ndparticle(n, l));
            if(ptable[p]->furry) {
                nf++;
            } else {
                vi = False;
                break;
            }
        }

        if(vi && nf%2 != 0) {
#if DBG
            if(dbg >= fldbg)
                printf("delete by Furry : v=%d\n", n);
#endif
            return False;
        }
    }
    return True;
}

/*==================================================================*/
Public  Void    fldump(NoArg)
/*--------------------------------------------------------------------
 *   Dump the table of f-lines. 
 */
{
    int f, n;

    printf("\n%% +++ %d fermion lines.\n", nflines);
    for(f = 0; f < nflines; f++) {
        printf("%% type=%s, length=%4d : ", 
               fltnam(fltable[f].type), fltable[f].length);
        for(n = 0; n < fltable[f].length; n++) {
            printf("[%d, %d] ",
                   tovert(fltable[f].ndlist[n]),
                   fltable[f].lglist[n]);
        }
        putchar(EOL);
    }
}

/*==================================================================*/
Private Void    flmswap(nd, l1, l2)
node    nd;
int     l1, l2;
/*--------------------------------------------------------------------
 *   Swap two legs l1 and l2 of node nd.
 */
{
    int t;

    t = ndnextn(nd, l1);
    ndnextn(nd, l1) = ndnextn(nd, l2);
    ndnextn(nd, l2) = t;

    t = ndparticle(nd, l1);
    ndparticle(nd, l1) = ndparticle(nd, l2);
    ndparticle(nd, l2) = t;

    t = ndiline(nd, l1);
    ndiline(nd, l1) = ndiline(nd, l2);
    ndiline(nd, l2) = t;

    t = ndilwk(nd, l1);
    ndilwk(nd, l1) = ndilwk(nd, l2);
    ndilwk(nd, l2) = t;
}

/*==================================================================*/
Public  Void    flmdir(NoArg)
/*--------------------------------------------------------------------
 *   Determine direction of i-lines of majorana.
 */
{
    int  f, n, lg[2], len, l;
    bool swp = False;
    node nd, ndo;
    pindex p;
    
    for(f = 0; f < nflines; f++) {
        len = fltable[f].length;
        if(fltable[f].type == FLOPEN)
            ndo = NDNULL;
        else
            ndo = fltable[f].ndlist[len-1];
        for(n = 0; n < len; n++) {
            /*  --<--lg[1][nd]lg[0]--<-- */

            nd   = fltable[f].ndlist[n];
            if(n == len-1 && fltable[f].type == FLOPEN)
                lg[0] = -1;
            else
                lg[0]  = fltable[f].lglist[n];
            if(ndo == NDNULL)
                lg[1] = -1;
            else
                lg[1] = floutlg(nd, ndo);
            for(l = 1; l >= 0; l--) {
                if(lg[l] < 0)
                    continue;
                p = abs(ndparticle(nd, lg[l]));
                if(ptable[p]->ptype == P_MAJORANA) {
                    p *= 1 - 2*l;
                    ndparticle(nd, lg[l]) = p;
                }
            }
            if(lg[0] >= 0 && lg[1] >= 0 && lg[0] < lg[1]) {
                flmswap(nd, lg[0], lg[1]);
                swp = True;
            }
            
            ndo = nd;
        }
    }
    if(swp)
        ilmake(False);

}

/*==================================================================*/
Public  int     flfact(NoArg)
/*--------------------------------------------------------------------
 *   Calculate sign of the graph for exchange of fermions,
 *   and (-1) for a fermion loop.
 */
{
    int  f, n, i1, i2, sgn, t;
    bool cont;
    int  *fsort;

    sgn = -1;

    if(nflines < 1)
        return sgn;

    if((fsort = (int*) malloc(2*nflines*sizeof(int))) == NULL) {
        error("flfact", "memory allocation failed");
        erfail();
    }

    n = 0;
    for(f = 0; f < nflines; f++) {
        if(fltable[f].type == FLOPEN) {
            fsort[n++] = fltable[f].ndlist[0];
            fsort[n++] = fltable[f].ndlist[fltable[f].length-1];
        } else {  /* loop */
            sgn = - sgn;
        }
    }

    /* sort and count sign */
    for(i1 = 0; i1 < n-1; i1++) {
        cont = True;
        while (cont) {
            cont = False;
            for(i2 = i1+1; i2 < n; i2++) {
                if(fsort[i1] > fsort[i2]) {
                    t         = fsort[i1];
                    fsort[i1] = fsort[i2];
                    fsort[i2] = t;
                    sgn *= -1;
                    cont = True;
                    break;
                }
            }
        }
    }

    free(fsort);

    return sgn;
}

/*==================================================================*/
Private Void    fldmpt(nl, typ)
int     nl, typ;
/*--------------------------------------------------------------------
 *   Dump temporary f-line.
 */
{
    int n;

    printf("\n%% +++ current fermion lines. length=%d, type=%s\n",
           nl, fltnam(typ));
    printf("%% ");
    for(n = 0; n < nl; n++) {
        printf(" [%d, %d] ",flnodes[n],fllegs[n]);
    }
    putchar(EOL);
}

/*==================================================================*/
Private int     flnextl(nd, lg)
node    nd;
int     lg;
/*--------------------------------------------------------------------
 *   Find which leg connecting to (nd, lg)
 */
{
    node ndn;
    int  l;

    ndn = ndnextn(nd, lg);
    for(l = 0; l < ndlegs(ndn); l++)
        if(ndnextn(ndn, l) == nd)
            return l;

    error("flnextl","illegal connextion");
    fprintf(Stderr, "nd=%d, lg=%, ndn=%d\n", nd, lg, ndn);
    erfail();
    return -1;
}

/*==================================================================*/
Private int     floutlg(nd, ndo)
node    nd, ndo;
/*--------------------------------------------------------------------
 *   Find the leg of nd connecting to ndo
 */
{
    int l;

    for(l = 0; l < ndlegs(nd); l++)
        if(ndnextn(nd, l) == ndo)
            return l;

    error("floutlg","illegal connextion");
    fprintf(Stderr, "nd=%d, ndo=%d\n", nd, ndo);
    erfail();
    return -1;
}
