/*********************************************************************
 *  File : matchv.c (prefix = pv)
 *
 *  match vertex
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "tokenh.h"

/*====================================================================
 *      Entry points
 */

/*      Find a node from the vertex table */
Public  int     pvable       ARG((node));
Public  int     pvcandp      ARG((node,pindex*));
Public  vindex  pvmatc       ARG((int,int*));
Public  vindex  pvmatch      ARG((int));
Public  int     pvmlist      ARG((node,node*));
Public  Void    pvreord      ARG((int));

/*  Global variables */

/* External references */
/* print.c */
Extern  Void    prdump       ARG((NoArg));
/* rdptbl.c */
Extern  vindex  pvaddct      ARG((vindex,int*));
Extern  Void    pvldump      ARG((FILE*, int, vindex[]));

/* Local functions */
Private int     pvwand       ARG((int,vindex[],int,vindex[]));
Private int     pvmat        ARG((int,vindex[],int*,pindex[],bool));

Local   vindex  pvwlist[LVTABLE];
Local   pindex  plist[MAXTL];
Local   int     mlist[MAXTL];

/*********************************************************************
 *  For assignment
 */

/*==================================================================*/
Private  int  pvwand(nvt, pvt, nvs, pvs)
int      nvt;
vindex   pvt[];
int      nvs;
vindex   pvs[];
/*--------------------------------------------------------------------
 * Construct and list of two list of vertices "pvs[nvs]" 
 * and "pvt[nvt]".
 * The resulting list is saved to "pvt[return_value]"
 * These list are assumed to be sorted in ascending order.
 *
 * Returns:
 *     the number of elements of the resulting list.
 *
 * Arguments:
 *     nvt : size of a list
 *     pvt : a list.  The resulting list is saved in there.
 *     nvs : size of another list
 *     pvs : another list.
 */
{
    int ls = 0, lt = 0, lr = 0;

#if DBG
    if(dbg >= pmdbg) {
        printf("pvwand : s%d :", nvs);
        pvldump(stdout, nvs, pvs);
        printf("pvwand : t%d :", nvt);
        pvldump(stdout, nvt, pvt);
    }
#endif

    /* enpty set */
    if(nvt < 1)
        return nvt;

    while(True) {
        while(ls < nvs && pvs[ls] < pvt[lt])
            ls++;
        if(ls >= nvs)
            break;

        while(lt < nvt && pvt[lt] < pvs[ls])
            lt++;
        if(lt >= nvt)
            break;

        if(pvs[ls] == pvt[lt]) {
            pvt[lr++] = pvt[lt];
            ls++;
            lt++;
        }
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("pvwand : r%d :", nvt);
        pvldump(stdout, nvt, pvt);
    }
#endif

    return lr;
}

/*==================================================================*/
Private int  pvlist(p, a, nl, pl)
pindex  p;
int     a;
int     nl;
vindex  pl[];
/*--------------------------------------------------------------------
 *  construct list of vertices which includes 
 *  particle "p"(particle a=0 or anti-particle a=1) in its "nl" legs.
 */
{
    int i, nvl, j;
    vindex v;

    j = 0;
    nvl = ptable[p]->nvlist[a];
    for(i = 0; i < nvl; i++) {
        v = ptable[p]->vlist[a][i];
        if(vtable[v]->nvlegs == nl) {
            if(j < LVTABLE)
                pl[j] = v;
            j++;
        }
    }
    p = (a == 0) ? p : -p;
    for(v = nvmodel; v < nvtable; v++) {
        if(vtable[v]->nvlegs != nl)
            continue;
        for(i = 0; i < nl; i++) {
            if(vtable[v]->vlegs[i] == p)
                break;
        }
        if(i < nl) {
            if(j < LVTABLE)
                pl[j] = v;
            j++;
        }
    }
    if(j >= LVTABLE){
        fprintf(Stderr, "too many candidate vertices %d(%d)\n",
                j, LVTABLE);
        exit(1);
    }
    return j;
}

/*==================================================================*/
Private int  pvmat(nd, pl, nt, pt, final)
int    nd;
vindex  pl[];
int    *nt;
pindex  pt[];
bool    final;
/*--------------------------------------------------------------------
 *  Construct the list of vertices which are assignable to the node
 *  "nd" ignoring multiplicity of the connection.
 *  This function return the length of the list.
 *
 *  input
 *    nd : node to be matched.
 *    nt : if(nt == NULL) then not use "nt" and "pt" for output.
 *    final : final check
 *
 *  output
 *    pl : list of possible vertices.
 *    pt : list of particles assignable to undefined legs.
 *         This data is prepared when input value of "nt" is not NULL.
 *    nt : length of "pt".
 *         This data is prepared when input value of "nt" is not NULL.
 */
{
#define MAXLGS 10

    bool match;
    int  l, a, nvl, nvl1, i, j, np=0, nl, lp;
    int  lt, v, vv, il;
    pindex p, vl[MAXLGS];
    pindex pl1[LVTABLE];

#if DBG
    if(dbg >= pmdbg) {
        printf("*dbg* pvmat(%d,pl,%p,pt,final)\n",nd,nt);
    }
#endif

    if(nt != NULL)
        *nt = 0;

    nl = ndlegs(nd);
    lp = ndloop(nd);

    if(nl > MAXLGS) {
        fprintf(Stderr, "*** pvmat : too small size of nl=%d\n", nl);
        exit(16);
    }

    /* construct list of assigned particles for "nd"
     * plist : list of particles,
     * mlist : list of multiplicity
     * np    : the length of the list
     */
    for(l = 0; l < ndlegs(nd); l++) {
        p = ndparticle(nd, l);
        if(p == PNULL) {
            if(final)
                /* final with not assigned leg */
                return 0;
            continue;
        }

        /* multiple connection ? */
        for(i = 0; i < np; i++) {
            if(plist[i] == p) {
                mlist[i]++;
                goto end_save;
            }
        }

        /* new particle */
        plist[np] = p;
        mlist[np] = 1;
        np++;

        end_save:;
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("pvmat : np = %d, ", np);
        for(l = 0; l < np; l++)
            printf("%d[%d] ", plist[l], mlist[l]);
        printf("\n");
    }
#endif

    /* construct list of candidate vertices
     * pl  : list of vertices
     * nvl : length of "pl".
     */
    nvl = 0;
    for(l = 0; l < np; l++) {
        p = plist[l];
        if(p == PNULL)  /* cannot occur ?? */
            continue;

        /* a = 0 for particle, 1 for anti-particle */
        a = (p < 0) ? 1 : 0;
        p = abs(p);

        /* construct initial list */
        if(l == 0) {

            nvl = pvlist(p, a, nl, pl);

        /* and with another list */
        } else {
#if DBG
            if(dbg >= pmdbg) {
                printf("pvmat : p=%d, a=%d :", p, a);
                pvldump(stdout, ptable[p]->nvlist[a], 
                                ptable[p]->vlist[a]);
            }
#endif
            nvl1 = pvlist(p, a, nl, pl1);
            nvl = pvwand(nvl, pl, nvl1, pl1);

            /* empty */
            if(nvl < 1)
                return 0;
        }
#if DBG
        if(dbg >= pmdbg) {
            printf("pvmat : l=%d, nvl=%d :", l, nvl);
            pvldump(stdout, nvl, pl);
        }
#endif
    }

    /* check each vertex
     * pl : list of vertices possible to assign
     * lt : length of the list.
     */
    lt = 0;

    /* for vv = candidat_vertex */
    for(v = 0; v < nvl; v++) {
        vv = pl[v];

        if(nl != vtable[vv]->nvlegs)
            continue;

        /* for counter term,  lp > vtable[vv]->loop */
        if(lp <  vtable[vv]->loop)
            continue;

        for(i = 0; i < nl; i++)
            vl[i] = vtable[vv]->vlegs[i];

        il = nl;       /* # of partices not in plist, mlist */

        /* for plist and multiplicity
         * vl : list of particle possible to assign.
         * il : length of vl.
         */
        for(l = 0; l < np; l++) {
            for(j = 0; j < mlist[l]; j++) {

                /* delete particle from vl */
                match = False;
                for(i = 0; i < il; i++) {
                    if(vl[i] == plist[l]) {
                        vl[i] = vl[--il]; 
                        match = True;
                        break;
                    }
                }
                if(!match)
                    goto not_match;  /* not accept */
            }
        }

        /* accept the vertex vv */
        pl[lt++] = vv;

        /* save possible particles which are assignable to
         * the undefined legs
         */
        if(nt != NULL) {
            for(i = 0; i < il; i++) {
                for(j = 0; j < *nt; j++)
                    if(pt[j] == vl[i])
                        break;
                if(j >= *nt)
                    pt[(*nt)++] = vl[i];
            }
        }
        not_match:;
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("pvmat nd = %d, nvl=%d, pvwlist=", nd, nvl);
        pvldump(stdout, nvl, pl);
    }
#endif
    return lt;
}

/*==================================================================*/
Public  Void pvlswap(nd, lg1, lg2)
int    nd;
int    lg1, lg2;
/*--------------------------------------------------------------------
 * swap two legs "lg1" and "lg2" of the node "nd".
 *
 * Arguments:
 *     nd : node.
 *     lg1, lg2 : legs to be swaped.
 */
{
    int lgn, lgp, lgi;

    lgn = ndnextn(nd, lg1);
    lgp = ndparticle(nd, lg1);
    lgi = ndiline(nd, lg1);

    ndnextn(nd, lg1)    = ndnextn(nd, lg2);
    ndparticle(nd, lg1) = ndparticle(nd, lg2);
    ndiline(nd, lg1)    = ndiline(nd, lg2);

    ndnextn(nd, lg2)    = lgn;
    ndparticle(nd, lg2) = lgp;
    ndiline(nd, lg2)    = lgi;
}

/*==================================================================*/
Public  Void pvreord(nd)
int    nd;
/*--------------------------------------------------------------------
 * Reorder the legs of node "nd" in accordance with the definition
 * in the model specification.
 * We assume vertex is already determined.
 *
 * Argument:
 *     nd : node.
 */
{
    int    iv, ig, nl;
    pindex vp, np;
    vindex v;

    v  = ndvertex(nd);
    nl = ndlegs(nd);
    for(iv = 0; iv < nl-1; iv++) {
        vp = vtable[v]->vlegs[iv];
        np = ndparticle(nd, iv);
        if(vp == np)
            continue;

        /* find "vp" from the latter part of the list and swap */
        for(ig = iv+1; ig < nl; ig++) {
            if(vp == ndparticle(nd, ig)) {
                pvlswap(nd, ig, iv);
                break;
            }
        }

        /* "vp" is not found */
        if(ig >= nl) {
            fprintf(Stderr, "pvreord: graph %d : ", cgraph->grid);
            fprintf(Stderr, "particle %d of ", vp);
            fprintf(Stderr, "node %d is not found ", nd);
            fprintf(Stderr, "in the vertex %d\n", v);
            exit(1);
        }
    }
}

/*==================================================================*/
Public  int     pvable(nd)
int    nd;
/*--------------------------------------------------------------------
 *  Check if this node is possible to assign a vertex.
 *  The node may have undetermined legs.
 *
 *  returns :
 *     the length of list of candidate vertices.
 *
 *  Argument:
 *     nd : node
 */
{
    int nvl;

    /* not construct the list of assignable particles */
    nvl = pvmat(nd, pvwlist, (int*)NULL, (pindex*)NULL, False);

#if DBG
    if(dbg >= pmdbg) {
        printf("pvable: nd = %d, nvl = %d\n", nd, nvl);
    }
#endif

    return nvl;
}

/*==================================================================*/
Public  vindex  pvmatch(nd)
int    nd;
/*--------------------------------------------------------------------
 * Find vertex in the model which is in the same structure of 
 * node "nd".
 * This function should be called for a node whithou undetermined
 * legs.
 * We assume only one vertex is possible to assing (bat condition???)
 *
 * Returns :
 *     vertex which should be assigned to the node.
 *     If failed to find then returns VNULL
 *
 * Argument :
 *     nd : node.
 */
{
#if DBG
    if(dbg >= pmdbg) {
        printf("pvmatch: nd = %d\n", nd);
    }
#endif

    /* final check withou constructing the list of assignable 
     * particles
     */
    if(pvmat(nd, pvwlist, (int*)NULL, (pindex*)NULL, True) > 0)
        return pvwlist[0];

    /* not found */
    else
        return VNULL;
}

/*==================================================================*/
Public  int     pvmlist(nd, nl)
int    nd;
int    nl[];
/*--------------------------------------------------------------------
 * Find vertex in the model which is in the same structure of 
 * node "nd".
 * This function should be called for a node whithou undetermined
 * legs.
 * We assume only one vertex is possible to assing (bat condition???)
 *
 * Returns :
 *     the number of candidate vertices.
 *
 * Argument :
 *     nd : node.
 *     nl : output : list of vertices
 */
{
    int lnl;

#if DBG
    if(dbg >= pmdbg) {
        printf("pvmlist: nd = %d\n", nd);
    }
#endif

    /* final check withou constructing the list of assignable 
     * particles
     */
    lnl = pvmat(nd, nl, (int*)NULL, (pindex*)NULL, True);

#if DBG
    if(dbg >= pmdbg) {
        printf("pvmlist: nd = %d, lnl=%d\n", nd, lnl);
    }
#endif
    return lnl;
}

/*==================================================================*/
Public  int pvcandp(nd, prtcl)
int    nd;
pindex *prtcl;
/*--------------------------------------------------------------------
 *  Construct the list of particles assignable to 
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      length of the list.
 *
 *  input
 *    nd : node
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    int np;

    Discard pvmat(nd, pvwlist, &np, prtcl, False);

    return np;
}

/*==================================================================*/
Public vindex pvmatc(nd, mord)
int    nd;
int    mord[];
/*--------------------------------------------------------------------
 *  Find vertex which match to node "nd".
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      vertex defined in the table
 *
 *  input
 *    nd : node
 *    mord : list of order of coupling constants.
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    int lnl, io, v, l;
    node nl[LVTABLE];

    lnl = pvmat(nd, nl, (int*)NULL, (pindex*)NULL, True);

    if(lnl < 1) {
        fprintf(Stderr, "pvmatc : cannot find counter term : node=%d\n");
        prdump();
        exit(1);
    }

    for(l = 0; l < lnl; l++) {
        v = nl[l];
        for(io = 0; io < nmordtbl; io++) {
            if(mord[io] != vtable[v]->morder[io])
                break;
        }
        if(l < lnl) {
            return v;
        }
    }

    return pvaddct(nl[0], mord);
}
