/*******************************************************************
 *    File : rdprocs.c
 *
 *    Read process
 *
 *    Calling sequence:
 *
 *      1. Initialization (read multiple process).
 *
 *              if(!rpinit(filename))
 *                  exit(FAIL);
 *
 *      2. Get next process.
 *
 *              while((nproc = rpread()) >= 0) {
 *                    ...
 *              }
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"
#include "tokenh.h"

/* Entry points */
Public  bool    rpinit       ARG((string));
Public  int     rpread       ARG((NoArg));
Public  Void    rprint       ARG((string));
Public  Void    rprin1       ARG((FILE*, int, bool));
Public  Void    rphelp       ARG((FILE*));

/* External references */
/* elem.c */
Extern  Void    elgene       ARG((pindex, int));
Extern  Void    elinit       ARG((NoArg));
/* ergrf.c */
Extern  Void    erdbg        ARG((string, string, int));
/* rdptbl.c */
Extern  pindex  pvfind       ARG((string));
Extern  Void    pvfname      ARG((string,string,int));
Extern  string  pvname       ARG((pindex));
Extern  Void    pvread       ARG((string));

/* Private functions */
Private bool    rpget        ARG((NoArg));
Private bool    rpexternal   ARG((int,int,int));
Private Void    rpdup        ARG((int,int));
Private Void    rpchk        ARG((NoArg));
Private Void    rpdump       ARG((NoArg));

/* Local variable */
Local   FILE    *rpfp;
Local   int     rpind;
Local   int     rpmord[MAXPROCS][MAXMORD]; 
Local   char    rpmdl[MAXPATHLEN];
                /* order of coupling constants */

/* help message */
Local string syntax[] = {
"Process file",
"  Syntax :",
"",
"\t file   ::= model-name [ order-definition ",
"\t\t | input-definition  ",
"\t\t | output-definition ",
"\t\t | \"opi=\" boolen \";\" ",
"\t\t | \"expand=\" boolean \";\" ",
"\t\t | \"block=\" boolen \";\" ",
"\t\t | \"tadpole=\" boolean \";\" ",
"\t\t | \"extself=\" boolean \";\" ",
"\t\t | \"selfe=\" boolean \";\" ",
"\t\t | \"countert=\" boolean \";\" ",
"\t\t | \"gm2=\" boolean \";\" ",
"\t\t | \"kinem=\" string \";\" ",
"\t\t ]* \"Pend\" \";\".",
"",
"\t order-definition ::= ",
"\t\t coupling-name \"= integer | ",
"\t\t                   \"{\" integer [ \",\" integer ]* \"};\".",
"",
"\t  input-definition ::= name | \"{\" name [ \",\" name ]* \"};\".",
"\t output-definition ::= name | \"{\" name [ \",\" name ]* \"};\".",
"",
"\t boolean = \"yes\" | \"no\"." };

/*==================================================================*/
Public  bool    rpinit(fn)
string  fn;
/*--------------------------------------------------------------------
    read processes
--------------------------------------------------------------------*/
{
    FILE*  echo;
    int    arg;
    trdsava sarea;

#if DBG
    erdbg("rpinit", "begin", rpdbg);
#endif

    /* echo : file pointer for echo of input. */
#if DBG
    echo = stdout;
#else
    echo = (FILE *) NULL;        /* no echo */
#endif

    /* allocate are for cprocess */
    if((cprocess = (tproc*) malloc(sizeof(tproc))) == NULL) {
        fprintf(Stderr, "\n*err* rpinit: memory allocation failed\n");
        exit(1);
    }

    /* open input file */
    if((rpfp = fopen(fn, "r")) == NULL) {
        fprintf(Stderr, "\n*err* rpinit: open error\n");
        fprintf(Stderr, "fname = %s\n", fn);
        return False;
    }

    /* initialize token package */
    rdinit(rpfp, RDNOCHANGE, echo);

    /* input : model="<model file>"; */
    rpind = rdstat();
    arg   = rdarg(rpind);
    fnmdl = rpmdl;
    if(rdchkid(rpind, "model") && rdtype(arg) == RDSTRING) {
        putchar('\n');
        pvfname(fnmdl, rdid(arg), MAXPATHLEN); /* model file name */
        rdsave(&sarea);      /* save file to use token for model */
        pvread(fnmdl);       /* read model file */
        rdrestore(&sarea);   /* recover file in the token package */
    } else {
        fprintf(Stderr, "\n*err* rpinit: model = \"<filename>\";");
        fprintf(Stderr, " is expected.\n");
        rdprint(rpind);
        rphelp(Stderr);
        return False;
    }

    /* the number of processes */
    rproc = 0;
    while (rpget()) /* read all processes into the table */
        ;

    rdend();
    fclose(rpfp);

    rpchk();

#if DBG
    if(dbg >= rpdbg) {
        rpdump();
        erdbg("rpinit", "end", rpdbg);
        for (arg = 0; arg < rproc; arg++)
            rprin1(stdout, arg, True);
    }
#endif

    return True;
}

/*==================================================================*/
Private bool    rpget(NoArg)
/*--------------------------------------------------------------------
    read one process
--------------------------------------------------------------------*/
{
    bool err, value;
    int  arg, found, prc, noprc, np, i;

#if DBG
    erdbg("rpget", "begin", rpdbg);
#endif

    /* process; */
    err = False;
    if((rpind = rdstat()) == RDEOF) {
        return False;
    } else if(rdtype(rpind) != RDID) {
        fprintf(Stderr, "\n*err* rpget: \"process;\" expected (not id)\n");
        exit(1);
    } else if(!rdchkid(rpind, "process")) {
        fprintf(Stderr, "\n*err* rpget: \"process;\" expected\n");
        exit(1);
    }

    /* check size of process table */
    if(rproc >= MAXPROCS - 1) {
        fprintf(Stderr, "\n*err* rpget: too many processes.\n");
        exit(1);
    }

    /* get space */
    np = rproc++;
    if((rpextn[np] = (rpextt*) malloc(sizeof(rpextt))) == NULL) {
        fprintf(Stderr, "\n*err* rpget: memory allocation error.\n");
        exit(1);
    }

    /* clear items and set by default values */
    rpextn[np]->nextern  = 0;
    rpextn[np]->expand   = True;
    rpextn[np]->onepi    = True;
    rpextn[np]->block    = False;
    rpextn[np]->tadpole  = False;
    rpextn[np]->extself  = False;
    rpextn[np]->selfe    = False;
    rpextn[np]->countert = True;
    rpextn[np]->gm2      = False;
    rpextn[np]->kinem    = NULL;
    rpextn[np]->cptest   = 0;

    /* prepare to accept more than one processes defined by the list
       of values of a coupling constant.    ex: qed={4, 6};
    */
    noprc = 0;   /* the number of processes */
    for(prc = 0; prc < MAXPROCS; prc++)
        for(i = 0; i < MAXMORD; i++)
            rpmord[prc][i] = 0;  /* buffer of coupling constants */

    /* read "pend;", "initial=...", "final=...", "opi=yes/no;", ... */
    while (True) {
        rpind = rdstat();
        arg = rdarg(rpind);

        if(rdtype(rpind) == RDEOF) {
            fprintf(Stderr, "\n*err* rpget: EOF before Pend\n");
            err = True;

        } else if(rdtype(rpind) != RDID) {
            fprintf(Stderr, "\n*err* rpget: illegal statement\n");
            rdprint(rpind);
            err = True;

        /* pend; */
        } else if(rdstrcmp(rdid(rpind), "pend") == 0) {
            break;

        /* initial=...*/
        } else if(rdstrcmp(rdid(rpind), "initial") == 0) {
            err |= !rpexternal(arg, np, ELINITIAL);

        /* final=... */
        } else if(rdstrcmp(rdid(rpind), "final") == 0) {
            err |= !rpexternal(arg, np, ELFINAL);

        /* opi=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "opi") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: opi should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->onepi = value;

        /* gm2=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "gm2") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: gm2 should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->gm2 = value;

        /* expand=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "expand") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: expand should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->expand = value;

        /* block=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "block") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: block should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->block = value;

        /* tadpole=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "tadpole") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: tadpole should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->tadpole = value;

        /* extself=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "extself") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: extself should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->extself = value;

        /* selfe=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "selfe") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: selfe should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->selfe = value;

        /* countert=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "countert") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: countert should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->countert = value;

        /* cptest=integer; */
        } else if(rdstrcmp(rdid(rpind), "cptest") == 0) {
            if(!rdintarg(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->cptest = value;

        } else if(rdstrcmp(rdid(rpind), "kinem") == 0) {
            if(rdtype(arg) == RDSTRING){
                rpextn[np]->kinem =
                    (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(rpextn[np]->kinem == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(rpextn[np]->kinem, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rpind);
                err = True;
            }

        /* coupling constants */
        /* xxx={number1, number2, ... }; */
        } else {
            found = False;
            arg = -1;
            for (i = 0; i < nmordtbl; i++) { /* find name of coupl. */
                if(rdstrcmp(mordtbl[i], rdid(rpind)) != 0)
                    continue;

                /* coupling constant is found*/
                found = True;

                /* for elements in list */
                for(arg = rdarg(rpind), prc = 0; arg >= 0;
                         arg = rdarg(arg), prc++) {
                    if(prc >= MAXPROCS) {
                        fprintf(Stderr, "\n*err* rpget: too many processes.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    }

                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "\n*err* rpget: illegal fname.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    } else if (err) {
                        break;
                    } else if (rpmord[prc][i] != 0) {
                        fprintf(Stderr, "\n*err* rpget: duplicate coupling.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    } else {
                        rpmord[prc][i] = rdint(arg);
                    }

                } /* for arg */

                if(!err && noprc > 0 && noprc != prc) {
                    fprintf(Stderr, "\n*err* rpget: different number");
                    fprintf(Stderr, " of processes.\n");
                    fprintf(Stderr, "Lengths of the list of coupling ");
                    fprintf(Stderr, "constant should be equal.\n");
                    rdprint(rpind);
                    err = True;
                }
                noprc = prc;

                break;
            } /* for coupling */

            if (!found) {
                fprintf(Stderr, "\n*err* rpget: illegal data.\n");
                rdprint(rpind);
                err = True;
            }
        } /* else : end of analysis of a statement */

    } /* while(True) : statements */

    /* check coupling constants */
    if (noprc < 1) {
        fprintf(Stderr, "\n*err* rpget: order of coupling constants ");
        fprintf(Stderr, " is not defined.\n");
        err = True;
    }

    /* check the number of coupling constants */
    if (rpextn[np]->nextern < 1) {
        fprintf(Stderr, "\n*err* rpget: no external particles\n");
        err = True;
    }

    if(err) {
        rdprint(rpind);
        rphelp(Stderr);
        exit(1);
    }

    /* copy order of coupling consts */
    rpdup(np, noprc);    

    /* check value ??? see rpread */

#if DBG
    erdbg("rpget", "end", rpdbg);
#endif

    return True;
}

/*==================================================================*/
Private bool    rpexternal(arg, np, iorf)
int arg;   /* arg of the statement */
int np;    /* process number */
int iorf;  /* ELINITIAL or ELFINAL */
/*--------------------------------------------------------------------
    read external particle.
--------------------------------------------------------------------*/
{
    int  pt;
    bool st;

#if DBG
    erdbg("rpexternal", "begin", rpdbg);
#endif

    pt = rpextn[np]->nextern;

    /* for elements in the argument list */
    st = True;
    for( ; arg >= 0; arg = rdarg(arg)) {

        /* particle name */
        if(rdtype(arg) != RDID) {
            fprintf(Stderr, "\n*err* rpexternal: illegal particle name.\n");
            rdprint(rpind);
            st = False;
            break;
        }

        if(pt >= MAXELINE - 1) {
            fprintf(Stderr, "\n*err* rpexternal:");
            fprintf(Stderr, " too many external particles.\n");
            st = False;
            break;
        }

        /* save 
           (if particle is not defined, then program stops in pvfind)
        */
        rpextn[np]->particle[pt] = pvfind(rdid(arg));
        rpextn[np]->iorf[pt] = iorf;
            
        pt++;
    }

    /* the number of external particles */
    rpextn[np]->nextern = pt;

#if DBG
    erdbg("rpexternal", "end", rpdbg);
#endif

    return st;
}

/*==================================================================*/
Private Void    rpdup(np, nprc)
int np;
int nprc;
/*--------------------------------------------------------------------
    save table of coupling constants
    They are saved into from np-th to (np+nprc-1)-th processes.
    From (np+1)-th to (np+nprc-1)-th processes, other information
    is duplicated from np-th process.
--------------------------------------------------------------------*/
{
    int i, p, n;

#if DBG
    erdbg("rpdup", "begin", rpdbg);
#endif

    for(i = 0; i < nprc; i++) {

        /* from (np+1)-th to (np+nprc-1)-th process */
        if(i > 0) {

            /* check table size */
            if(rproc > MAXPROCS) {
                fprintf(Stderr, "\n*err* rpdup: too many processes.\n");
                exit(1);
            }

            /* get space */
            n = rproc++;
            if((rpextn[n] = (rpextt*) malloc(sizeof(rpextt)))
                     == NULL) {
                fprintf(Stderr, "\n*err* rpdup: memory allocation error\n");
                exit(1);
            }

            /* copy data */
            rpextn[n]->nextern  = rpextn[np]->nextern;
            rpextn[n]->expand   = rpextn[np]->expand;
            rpextn[n]->onepi    = rpextn[np]->onepi;
            rpextn[n]->block    = rpextn[np]->block;
            rpextn[n]->tadpole  = rpextn[np]->tadpole;
            rpextn[n]->extself  = rpextn[np]->extself;
            rpextn[n]->selfe    = rpextn[np]->selfe;
            rpextn[n]->countert = rpextn[np]->countert;
            rpextn[n]->cptest   = rpextn[np]->cptest;
            rpextn[n]->gm2      = rpextn[np]->gm2;
            rpextn[n]->kinem    = rpextn[np]->kinem;

            /* copy external particles */
            for(p = 0; p < rpextn[np]->nextern; p++) {
                rpextn[n]->particle[p] = rpextn[np]->particle[p];
                rpextn[n]->iorf[p]     = rpextn[np]->iorf[p];
            }

        /* np-th process */
        } else {
            n = np;
        }

        /* copy coupling constants */
        for(p = 0; p < nmordtbl; p++)
            rpextn[n]->morder[p] = rpmord[i][p];
    }

#if DBG
    erdbg("rpdup", "end", rpdbg);
#endif

}

/*==================================================================*/
Public  int     rpread(NoArg)
/*--------------------------------------------------------------------
    read the next process from the process table.
--------------------------------------------------------------------*/
{
    static int nproc = 0, np;
    int lp2, i;

#if DBG
    erdbg("rpread", "begin", rpdbg);
#endif


    if(rproc <= nproc) { /* out of the process table */
#if DBG
        if(dbg >= rpdbg)
            printf("*dbgr* pread :rproc = %d, nproc = %d\n", 
                   rproc, nproc);
#endif

        return -1;
    }

    /* new process number */
    np = rprocc = nproc++;

    /* copy as the current process */
    nextern = rpextn[np]->nextern;   
             /* the number of external particles */

    order = 0;                       /* total order of coupling */
    for(i = 0; i < nmordtbl; i++) {
        morder[i] = rpextn[np]->morder[i];
        order    += morder[i];
    }

    lp2 = order - nextern + 2;       /* 2*loop */
    if(lp2 % 2 != 0) {
        fprintf(Stderr, "\n*err* rpread: illegal order.\n");
        rprin1(Stderr, np, True);
        exit(1);
    }
    loop = lp2 / 2;
    if(loop < 0) {
        fprintf(Stderr, "\n*err* rpread: negative number of loops.\n");
        rprin1(Stderr, np, True);
        exit(1);
    }

    /* flags */
    expand   = rpextn[np]->expand;
    onepi    = rpextn[np]->onepi;
    block    = rpextn[np]->block;
    tadpole  = rpextn[np]->tadpole;
    extself  = rpextn[np]->extself;
    selfe    = rpextn[np]->selfe;
    countert = rpextn[np]->countert;
    cptest   = rpextn[np]->cptest;
    gm2      = rpextn[np]->gm2;
    kinem    = rpextn[np]->kinem;

    /* normalization */
    if(loop < 1 || !expand) {
        onepi   = False;
        block   = False;
        tadpole = False;
        extself = False;
        selfe   = False;
        countert = False;
    } else {
        if(nextern < 2) {
            tadpole = True;
        } else if(nextern < 3) {
            extself = True;
            selfe   = True;
        }
    }

    /* initialize table of graph elements */
    elinit();

    /* enter external particles */
    for(i = 0; i < nextern; i++) {
        elgene(rpextn[np]->particle[i], rpextn[np]->iorf[i]);
    }

#if DBG
    erdbg("rpread", "end", rpdbg);
#endif

    /* construct current process data */
    cprocess->procid   = nproc;
    cprocess->graphs   = 0;        /* initial value */
    cprocess->nextern  = nextern;
    cprocess->loop     = loop;
    cprocess->order    = order;
    cprocess->bgraph   = 0;
    cprocess->maxnodes = MAXNODES;
    cprocess->maxlegs  = MAXLEGS;
    cprocess->msize    = sizeof(tgraph)
                + sizeof(tnode) * cprocess->maxnodes
                + sizeof(tleg)  * cprocess->maxlegs;
    cprocess->expand   = expand;
    cprocess->onepi    = onepi;
    cprocess->gm2      = gm2;
    cprocess->kinem    = kinem;

    cprocess->mult     = PC_NORMAL;
    cprocess->proc1    = 0;
    cprocess->proc2    = 0;
    for(i = 0; i < nmordtbl; i++)
        cprocess->morder[i] = morder[i];

    return nproc;
}

/*==================================================================*/
Public  Void    rprint(fn)
string  fn;
/*--------------------------------------------------------------------
    Print the processes table in the form of process file.
--------------------------------------------------------------------*/
{
    FILE *fp;
    int  np;
    bool mode;

    if((fp = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* rprint: open error to write ");
        fprintf(Stderr,"filename = %s\n", fn);
        exit(1);
    }

    mode = True;
    for(np = 0; np < rproc; np++) {
        rprin1(fp, np, mode);
        mode = False;
    }
    fclose(fp);
}

/*==================================================================*/
Public  Void    rprin1(fp, proc, mdl)
FILE *fp;
int proc;
bool mdl;  /* with model file name ??? */
/*--------------------------------------------------------------------
    Print the processes.
--------------------------------------------------------------------*/
{
    int i, j;

    /* name of model file */
    if(mdl) {
        fprintf(fp, "%%\n");
        fprintf(fp, "Model=\"%s.mdl\";\n", fnmdl);
    }

    /* process number */
    fprintf(fp, "%%\n");
    fprintf(fp, "Process; %% %d\n", proc);

    /* order of coupling constants */
    for(i = 0; i < nmordtbl; i++)
        fprintf(fp, "%s=%d; ", mordtbl[i], rpextn[proc]->morder[i]);
    fprintf(fp,"\n");

    /* initail particles */
    j = 0;
    fprintf(fp, "Initial={");
    for(i = 0; i < rpextn[proc]->nextern; i++)
        if(rpextn[proc]->iorf[i] == ELINITIAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            fprintf(fp, "%s", pvname(rpextn[proc]->particle[i]));
        }
    fprintf(fp,"};\n");

    /* final particles */
    j = 0;
    fprintf(fp, "Final  ={");
    for(i = 0; i < rpextn[proc]->nextern; i++)
        if(rpextn[proc]->iorf[i] == ELFINAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            fprintf(fp, "%s", pvname(rpextn[proc]->particle[i]));
        }
    fprintf(fp,"};\n");

    /* flags */
    fprintf(fp,"Expand=%d; Opi=%d; Block=%d; Gm2=%d\n",
            rpextn[proc]->expand, rpextn[proc]->onepi, 
            rpextn[proc]->block, rpextn[proc]->gm2);
    fprintf(fp,"Tadpole=%d; Extself=%d; Selfe=%d; Countert=%d\n",
            rpextn[proc]->tadpole, rpextn[proc]->extself,
            rpextn[proc]->selfe,   rpextn[proc]->countert);
    if(rpextn[proc]->kinem != NULL)
        fprintf(fp,"kinem=%s\n", rpextn[proc]->kinem);

    /* end */
    fprintf(fp, "Pend;\n");
}

/*==================================================================*/
Public  Void    rphelp(fp)
FILE    *fp;
/*--------------------------------------------------------------------
    Print help
--------------------------------------------------------------------*/
{
    int i, synlen;

    synlen = sizeof(syntax)/sizeof(syntax[0]);
    for(i = 0; i < synlen; i++)
        fprintf(fp, "%s\n", syntax[i]);
}

/*==================================================================*/
Private Void    rpchk(NoArg)
/*--------------------------------------------------------------------
   Check consistency.
--------------------------------------------------------------------*/
{
    int np, i, lp2, lp, ordr, fn, ne, p, pt, a;
    bool err=False;

    for(np = 0; np < rproc; np++) {
        ne = rpextn[np]->nextern;   
    
        ordr = 0;                       /* total order of coupling */
        for(i = 0; i < nmordtbl; i++) {
            ordr += rpextn[np]->morder[i];
        }
    
        lp2 = ordr - ne + 2;       /* 2*loop */
        if(lp2 % 2 != 0) {
            fprintf(Stderr, "** err **: illegal order.\n");
            rprin1(Stderr, np, True);
            err = True;
        }
        lp = lp2 / 2;
        if(lp < 0) {
            fprintf(Stderr, "** err **: negative number of loops.\n");
            rprin1(Stderr, np, True);
            err = True;
        }

        fn = 0;
        for(p = 0; p < rpextn[np]->nextern; p++) {
            pt = rpextn[np]->particle[p];
            if(pt < 0) {
                pt = - pt;
                a  = -1;
            } else {
                a  = 1;
            }
            switch(ptable[pt]->ptype) {
                case P_FERMION:
                case P_GHOST:
                    if(majop)
                        fn++;
                    else if(rpextn[np]->iorf[p] == ELFINAL)
                        fn -= a;
                    else
                        fn += a;
                    break;
                case P_MAJORANA:
                    fn++;
                    break;
            }
        }
        if((fn != 0 && !majop) || (fn%2 != 0 && majop)) {
            fprintf(Stderr, "** err **: illegal fermion number\n");
            rprin1(Stderr, np, True);
            err = True;
        }
    }
    if(err)
        exit(1);
}

/*==================================================================*/
Private Void    rpdump(NoArg)
/*--------------------------------------------------------------------
   Dump the process table.
--------------------------------------------------------------------*/
{
    int p, i;

    printf("rpdump : dump process table: rproc = %d\n", rproc);

    for(p = 0; p < rproc; p++) {
        printf("%d-th process\n", p);

        printf("nextern=%d, expand=%d, ounepi=%d, block=%d, ",
            rpextn[p]->nextern,
            rpextn[p]->expand,
            rpextn[p]->onepi,
            rpextn[p]->block);
        printf("tadpole=%d, extself=%d, selfe=%d, countert=%d, gm2=%d\n",
            rpextn[p]->tadpole,
            rpextn[p]->extself,
            rpextn[p]->selfe,
            rpextn[p]->countert,
            rpextn[p]->gm2);
        if(rpextn[p]->kinem != NULL)
            printf("kinem=%s\n", rpextn[p]->kinem);

        for(i = 0; i < nmordtbl; i++)
            printf("%s=%d ", mordtbl[i], rpextn[p]->morder[i]);
        putchar(EOL);

        for(i = 0; i < rpextn[p]->nextern; i++)
            printf("%4d %s %s(%d)\n", i, eliorf(rpextn[p]->iorf[i]),
                   pvname(rpextn[p]->particle[i]),
                   rpextn[p]->particle[i]);
    }
}
