/*********************************************************************
 *  File : cfconv.c
 *
 *  Convert graph data into cfbase format
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "glue4.h"
#include "cfbase.h"

/*
#define OLDINDEX
*/

/*--------------------------------------------------------------------
 * Entry points
 */
Public  Void    cfcbase      ARG((cftbase*));
Public  int     cfcproj      ARG((cftbase*, cftgraph*));
Public  Void    cfcbfree     ARG((cftbase*));
Public  Void    cfcgfree     ARG((cftgraph*));
Public  int     cfnoind      ARG((int));
Public  int     cfeoind      ARG((int));

/*--------------------------------------------------------------------
 * External references
 */

/* Private functions
 */
Private int     cfckind      ARG((int));
Private Void    cfcbinit     ARG((cftbase*, int));
Private Void    cfcginit     ARG((cftgraph*, int, int, int));
Private int     cfcgl4       ARG((int*));
Private int     cfcgl3s      ARG((int*));
Private int     cfvkind      ARG((int, int*));
Private int     cfnindex     ARG((int));
Private int     cfeindex     ARG((int));

/*--------------------------------------------------------------------
 * local variables
 */
/* particle code of an edge
 */
#define ilpartl(l,lg)     \
        ndparticle(iltable[l].ilnodes[lg],iltable[l].illegs[lg])
#define ilpart(l)     ilpartl(l, 1)

Local int nextern;    /* the number of external particles
                       */
Local int nvert0;     /* the number of original vertices
                       */
Local int nvert;      /* the number of vertices including expanded
                       * three-gluon vertices
                       */
Local int nilines;    /* the number of edges
                       */
Local int noffsv;     /* amount of shift for nodes for 
                       * index conversion
                       */
Local int noffse;     /* amount of shift for edges for 
                       * index conversion
                       */

/*==================================================================*/
Public  Void    cfcbase(cbase)
cftbase *cbase;
/*--------------------------------------------------------------------
 * 1. Initialize "cbase" allocating space for its components.
 *    Space for elements concerning about color base should be 
 *    allocated after nbase is determined.
 *
 * 2. Set information about external particles.
 *
 * Output:
 *    cbase : data to be initialized
 */
{
    int  p;
    node n, nd;

    nextern = cprocess->nextern;

    /* Constants for index conversion
     */
    noffsv  = nextern;       /* for vertices */
    noffse  = 0;             /* for edges */

    /* Initialize cbase data structure : allocation of memory.
     * Memory for variables related to color base should
     * be allocated in "cfbase.c".
     */
    cfcbinit(cbase, nextern);

    /* Fill type of external particles to "cbase->ctype"
     */
    for(n = 0; n < nextern; n++) {

        /* particle code of the external particle
         * +:particle,  -:ant-particle
         */
        p  = ndname(n);

        /* convert index
         */
        nd = - cfnindex(n);

        /* save particle type : CFQUARK etc.
         */
        cbase->cptype[nd] = cfckind(p);
    }
}

/*==================================================================*/
Private int     cfckind(p)
int     p;
/*--------------------------------------------------------------------
 * Convert particle code, defined in model file, to particle type 
 * for cfbase.
 *
 * Input
 *    p : particle code with sign (+:particle,  -:anti-particle)
 *
 * Returns
 *    particle type, that is, CFQUARK, etc.
 */
{
    int c;
    int sgn = 1;

    if(p < 0) {
        p   = - p;
        sgn = - 1;
    }

    /* dim. of color representation of the particle
     * We expect it one of 1, 3 or 8.
     * We assume
     *
     *         1 :                           : CFNOCOLOR
     *         3 : particle != anti-particle : CFQARK/CFAQUARK
     *         3 : particle == anti-particle : no such case
     *         8 : particle == anti-particle : CFGLUON
     *         8 : particle != anti-particle : CFGHOST/CFAGHOST
     */
    c = ptable[p]->color;
    switch (c) {

        case 3 :
            /* Anti-particle is defined in the model file
             */
            if(ptable[p]->antip < 0) {
                return ((sgn > 0) ? CFQUARK : CFAQUARK);

            /* We assume 3* representation is carried by
             * an anti-particle of a particle with 3 representation
             */
            } else {
                fprintf(Stderr, "*** cfckind : particle %s has no ",
                        ntable[ptable[p]->pname[0]].name);
                fprintf(Stderr, "anti-particle but color ");
                fprintf(Stderr, "representation = 3\n");
                prdump();
                exit(1);
            }

        case 8 :
            /* Anti-particle is defined in the model file
             */
            if(ptable[p]->antip < 0) {
                return ((sgn > 0) ? CFGHOST : CFAGHOST);

            /* Anti-particle is not defined in the model file
             * but it appears in a graph.
             */
            } else if(sgn < 0) {
                fprintf(Stderr, "*** cfckind : particle %s has no ",
                        ntable[ptable[p]->pname[0]].name);
                fprintf(Stderr, "anti-particle but is used for ");
                fprintf(Stderr, "representation = 8\n");
                prdump();
                exit(1);

            /* We assume particle which is equal to anti-particle and
             * is in 8 dim. representation is a gluon.
             */
            } else {
                return CFGLUON;
            }

        case 1 :
            return CFNOCOLOR;

        /* Illegal color representation */
        default :
            fprintf(Stderr, "*** cfckind : illegal color ");
            fprintf(Stderr, "representation=%d for particle %s\n",
                    c, ntable[ptable[p]->pname[0]].name);
            prdump();
            exit(1);
    }
}

/*==================================================================*/
Private Void    cfcbinit(cbase, next)
cftbase *cbase;
int     next;       /* the number of external particles */
/*--------------------------------------------------------------------
 * Allocate memory for the elements of "cbase" for "next" external
 * particles.
 */
{
    int i;

    /* the number of external particles */
    cbase->nex = next;

    /* int cptype[0..next] */
    if(next > 0) {
        cbase->cptype = (int *) malloc(sizeof(int)*(next+1));
        if(cbase->cptype == NULL) {
            fprintf(Stderr, "*** cfcbinit: memory allocation error\n");
            exit(1);
        }
        /* put default value */
        for(i = 1; i <= next; i++)
            cbase->cptype[i] = CFNOCOLOR;

    /* no external particles */
    } else {
        cbase->cptype = NULL;
    }

    /* Memory of elements related to the color base is not
     * allocated here.  It should be done in "cfbase.c"
     */
    cbase->nbase   = 1;
    cbase->nexcolor= 0;
    cbase->nquarkp = 0;
    cbase->ngluon  = 0;
    cbase->quark   = NULL;
    cbase->antiq   = NULL;
    cbase->gluon   = NULL;
    cbase->constr  = NULL;
    cbase->nconid  = NULL;
    cbase->conid   = NULL;
    cbase->cwgt    = NULL;
}

/*==================================================================*/
Public  Void    cfcbfree(cbase)
cftbase *cbase;
/*--------------------------------------------------------------------
 * free allocated memory area in "cbase".
 */
{
    int i;

    /* free cptype[] */
    if(cbase->cptype != NULL) {
        free(cbase->cptype);
        cbase->cptype = NULL;
    }

    /* free cwgt[0..nbase-1][] */
    if(cbase->cwgt != NULL) {
        /* free cwgt[i][] */
        for(i = 0; i < cbase->nbase; i++) {
            if(cbase->cwgt[i] != NULL) {
                free(cbase->cwgt[i]);
                cbase->cwgt[i] = NULL;
            }
        }
        /* free cwgt[] */
        free(cbase->cwgt);
        cbase->cwgt = NULL;
    }

    /* free quark[] */
    if(cbase->quark != NULL) {
        free(cbase->quark);
        cbase->quark = NULL;
    }

    /* free antiq[] */
    if(cbase->antiq != NULL) {
        free(cbase->antiq);
        cbase->antiq = NULL;
    }

    /* free gluon[] */
    if(cbase->gluon != NULL) {
        free(cbase->gluon);
        cbase->gluon = NULL;
    }
    
    /* free constr[] */
    if(cbase->constr != NULL) {
        free(cbase->constr);
        cbase->constr = NULL;
    }
    
    /* free nconid[] */
    if(cbase->nconid != NULL) {
        free(cbase->nconid);
        cbase->nconid = NULL;
    }

    /* free conid[0..nbase-1][] */
    if(cbase->conid != NULL) {
        /* free conid[i][] */
        for(i = 0; i < cbase->nbase; i++) {
            if(cbase->conid[i] != NULL) {
                free(cbase->conid[i]);
                cbase->conid[i] = NULL;
            }
        }
        /* free conid[] */
        free(cbase->conid);
        cbase->conid = NULL;
    }
}

/*==================================================================*/
Public  int     cfcproj(cbase, colgr)
cftbase  *cbase;
cftgraph *colgr;
/*--------------------------------------------------------------------
 * save data of a graph.
 * Returns the sign factor in terms of three-gluon vertices.
 */
{
    int i, il, p, n, ns, ne, nl, nd, ni, es, ee, l4, lg;
    int nk[CFPKIND];            /* work to determine vertex type */
    int nglue4 = 0;             /* the number of 4 gluon vertices */
    int *glue4 = NULL;          /* the list of 4 gluon vertices */
    int gl3il[3];               /* work to determine the sign of
                                 * a 3-gluon vertex
                                 */
    int sgn = 1;                /* the product of signatures
                                 * correspondent to the permutation
                                 * of legs of 3-gluon vertices
                                 */

    /* the number of original vertices */
    nvert0  = cgraph->nnodes - cprocess->nextern;

    /* glue4[0..nvert0] : table of four-gluon vertices.
     * The maximum possible number of such vertices are "nvert0"
     */
    if((glue4 = (int*) malloc((nvert0+1)*sizeof(int))) == NULL) {
        fprintf(Stderr, "*** cfcproj : memory allocation\n");
        exit(1);
    }

    /* the number of 4-gluon vertices
     */
    nglue4  = cfcgl4(glue4);

    /* "nglue4" nodes are added to the original graph in expanding 
     * four-gluon vertices
     */
    nvert   = nvert0 + nglue4;

    /* "nglue4" edges are added to the original graph in expanding 
     * four-gluon vertices.
     * Exclude the dummy edge at index=0 which is already included in 
     * "cgraph->nilines".
     */
    nilines = cgraph->nilines - 1 + nglue4;

    /* allocate area for "colgr"
     */
    cfcginit(colgr, nvert, nilines, cbase->nbase);

#if DBG
    if(dbg > 0)
        prdump();
#endif

    /* Fill tedge[0..nilines] and edge[0..1][0..nilines].
     * Edge 0 is a dummy.
     */
    for(i = 1; i < cgraph->nilines; i++) {
        p  = ilpart(i);              /* particle of the edge */
        ns = iltable[i].ilnodes[0];  /* starting node */
        ne = iltable[i].ilnodes[1];  /* ending node */

        il = cfeindex(i);    /* index for cfbase */

        if(p > 0) {                  /* save for cfbase */
            colgr->tedge[il]    = cfckind(p);
            colgr->edge[0][il]  = cfnindex(ns);
            colgr->edge[1][il]  = cfnindex(ne);
        } else {                     /* reverse direction */
            colgr->tedge[il]    = cfckind(-p);
            colgr->edge[0][il]  = cfnindex(ne);
            colgr->edge[1][il]  = cfnindex(ns);
        }
    }

    /* fill lvtx[0..nvtx] and nleg[0..nvtx].
     * Vertex 0 is a dummy.
     */
    for(n = cprocess->nextern; n < cgraph->nnodes; n++) {
        nd = cfnindex(n);    /* index for cfbase */

        /* Construct a table "nk" of the numbers of particles 
         * classified in their kinds.
         */
        for(i = 0; i < CFPKIND; i++)
            nk[i] = 0;

        if(ndgl4(n) > 0) {    /* four-gluon vertex */
            nl = 3;           /* save as a three-gluon vertex */
            nk[CFGLUON] = 3;
        } else {
            nl = ndlegs(n);   /* the number of legs */
            for(i = 0; i < nl; i++)
                nk[cfckind(ndparticle(n, i))]++;
        }

        colgr->lvtx[nd] = cfvkind(nl, nk);  /* kind of the vertex */
        colgr->nleg[nd] = nl;              /* the number of edges */

        /* Calculate the sign of original 3-gluon vertex.
         */
        if(colgr->lvtx[nd] == CFVGGG && ndlegs(n) == 3) {
            for(i = 0; i < 3; i++)
                gl3il[i] = cfeindex(ndiline(n, i));
            sgn *= cfcgl3s(gl3il);
        }
    }

    /* Index of the node to be added when 4-gluon vertices are
     * expanded (as an original index)
     */
    ne = cgraph->nnodes;
    /* Index of the edge to be added when 4-gluon vertices are
     * expanded (as an original index)
     */
    ni = cgraph->nilines;
    for(n = 0; n < nglue4; n++, ne++, ni++) {
        ns = glue4[n];             /* original four-gluon vertex */
        es = cfnindex(ns);         /* as an index of cfbase */

        /* create new 3 gluon vertex with index "ne" */
        ee = cfnindex(ne);                 /* as an index of cfbase */
        colgr->lvtx[ee] = CFVGGG;
        colgr->nleg[ee] = 3;

        /* create a new edge with index "ni" */
        il = cfeindex(ni);                 /* as an index of cfbase */
        colgr->tedge[il]    = CFGLUON;
        colgr->edge[0][il]  = es;
        colgr->edge[1][il]  = ee;

        /* reconnect edges which is connected to the original vertex
         * legs (l0, l1, l2, l3) ==> (l0, l1, ln) (l2, l3, ln)
         *                            original v.     new v.
         * ndgl4(ns) : 1..3 : corresponds to ndgl4(ns)-th 3-g. vertex
         * glue4leg[n][l] : permutation of the l-th leg corresponding to
         *                  the n-th 3-gluon vertex.
         */
        for(lg = 2; lg < 4; lg++) {
            l4 = glue4leg[ndgl4(ns)][lg];         /* leg to be moved */
            i  = ndiline(ns, l4);          /* edge to be reconnected */
            il = cfeindex(i);               /* as an index of cfbase */

            /* We assume the both sides a gluon propagator are different
             * vertices.
             * Is it correct for SUSY ???
             */
            if(colgr->edge[0][il] == es) {              /* es ==> ee */
                colgr->edge[0][il] = ee;
            } else if(colgr->edge[1][il] == es) {
                colgr->edge[1][il] = ee;
            } else {
                fprintf(Stderr, "*** cfcproj : inconsistent \n");
                prdump();
                exit(1);
            }
        }

        /* sign for "es" and "ee" */
        for(i = 0; i < 4; i += 2) {           /* i=0: es,  i=2: ee */
            for(lg = 0; lg < 2; lg++) {
                l4 = glue4leg[ndgl4(ns)][lg + i];
                gl3il[lg] = cfeindex(ndiline(ns, l4));
            }
            gl3il[2] = cfeindex(ni);    
            sgn *= cfcgl3s(gl3il);
        }
    }

    /* free work space */
    free(glue4);

    return sgn;
}

/*==================================================================*/
Private int     cfcgl3s(gl3)
int    *gl3;
/*--------------------------------------------------------------------
 * Returns the sign of permutation which sort gl3.
 */
{
    int t;
    int sgn = 1;

#if DBG
    if(dbg > 0)
        printf("cfcgl3: (%d, %d, %d) ", gl3[0], gl3[1], gl3[2]);
#endif

    /* sort of an array with 3 elements */
    if(gl3[0] > gl3[1]) {
        t = gl3[0];
        gl3[0] = gl3[1];
        gl3[1] = t;
        sgn *= -1;
    }
    if(gl3[1] > gl3[2]) {
        t = gl3[1];
        gl3[1] = gl3[2];
        gl3[2] = t;
        sgn *= -1;
    }
    if(gl3[0] > gl3[1]) {
        t = gl3[0];
        gl3[0] = gl3[1];
        gl3[1] = t;
        sgn *= -1;
    }
#if DBG
    if(dbg > 0)
        printf("--> %d\n", sgn);
#endif
    return sgn;
}

/*==================================================================*/
Private int     cfcgl4(gl4)
int    *gl4;
/*--------------------------------------------------------------------
 * count the number of four-gluon vertices and make a table of such 
 * vertices.
 */
{
    int nd, ng;

/*    for(nd = 0, ng = 0; nd < cgraph->nnodes; nd++) {*/
    for(nd = cprocess->nextern, ng = 0; nd < cgraph->nnodes; nd++) {
        if(ndgl4(nd) > 0) {
            gl4[ng++] = nd;
        }
    }
    return ng;
}

/*==================================================================*/
Private int     cfnindex(n)
int     n;
/*--------------------------------------------------------------------
 * convert index of a node for color base package
 *
 * returns
 *         -e, ..., -1 : external line  <= 0, ... , (e-1)
 *          1, ..., v  : vertex         <= e, ... , (v-e-1)
 *    where, e = #external_particles = noffsv and
 *    v = #vertices
 */
{
    if(n < noffsv) {         /* external particle */
#ifdef OLDINDEX
        return - 1 - n;
#else
        return n - noffsv;
#endif
    } else {               /* vertex */
        return n - noffsv + 1;
    }
}

/*==================================================================*/
Public  int     cfnoind(n)
int     n;
/*--------------------------------------------------------------------
 * convert index of a node from color base package to original one
 *
 * returns
 *         -e, ..., -1 : external line  => 0, ... , (e-1)
 *          1, ..., v  : vertex         => e, ... , (v-e-1)
 *    where, e = #external_particles = noffsv and
 *    v = #vertices
 */
{
    if(n < 0) {         /* external particle */
#ifdef OLDINDEX
        return - 1 - n;
#else
        return n + noffsv;
#endif
    } else {               /* vertex */
        return n + noffsv - 1;
    }
}

/*==================================================================*/
Private int     cfeindex(e)
int     e;
/*--------------------------------------------------------------------
 * convert index of an edge for color base package
 *
 * returns
 *        1, ..., i : edge            <= (e+1), ... ,(e+i)
 *        error                       <= 0         : dummy i-line
 *        error                       <= 1, ..., e : ext. part.
 *    where, e = #external_particles = noffse
 *
 */
{
    if(e <= noffse) {
        fprintf(Stderr, "*** cfeindex : illegal index of edge\n");
        exit(1);
    } else {
        return e - noffse;
    }
}

/*==================================================================*/
Public  int     cfeoind(e)
int     e;
/*--------------------------------------------------------------------
 * convert index of an edge from color base package to original one
 *
 * returns
 *        1, ..., i : edge            => (e+1), ... ,(e+i)
 *        error                       => 0         : dummy i-line
 *        error                       => 1, ..., e : ext. part.
 *    where, e = #external_particles = noffse
 *
 */
{
    return e + noffse;
}

/*==================================================================*/
Private Void    cfcginit(colgr, nv, ne, nb)
cftgraph *colgr;
int      nv;        /* the number of nodes */
int      ne;        /* the number of edges */
int      nb;        /* the number of color bases */
/*--------------------------------------------------------------------
 * Allocate memory of the elements of "colgr".
 */
{
    int i;

    colgr->nvtx  = nv;
    colgr->nedge = ne;

    if(nv > 0) {
        /* int lvtx[0..nv] */
        colgr->lvtx = (int *) malloc(sizeof(int)*(nv+1));

        /* int nleg[0..nv] */
        colgr->nleg = (int *) malloc(sizeof(int)*(nv+1));
        if(colgr->lvtx == NULL || colgr->nleg == NULL) {
            fprintf(Stderr, "*** cfcginit: memory allocation error\n");
            exit(1);
        }

        /* default values */
        for(i = 0; i < nv; i++) {
            colgr->lvtx[i] = CFVDEL;
            colgr->nleg[i] = 0;
        }

        /* int tedge[0..ne] */
        colgr->tedge   = (int *) malloc(sizeof(int)*(ne+1));

        /* int edge[0..1][0..ne] */
        colgr->edge[0] = (int *) malloc(sizeof(int)*(ne+1));
        colgr->edge[1] = (int *) malloc(sizeof(int)*(ne+1));
        if(colgr->tedge   == NULL ||
           colgr->edge[0] == NULL ||
           colgr->edge[1] == NULL)
        {
            fprintf(Stderr, "*** cfcginit: memory allocation error\n");
            exit(1);
        }

        /* default values */
        for(i = 0; i <= ne; i++) {
            colgr->tedge[i]   = CFNPUNDEF;
            colgr->edge[0][i] = 0;
            colgr->edge[1][i] = 0;
        }

        /* double wgt[0..nb-1] */
        /* (K.K.) colgr->wgt = (double *) malloc(sizeof(double)*nb); */
        colgr->wgt = (cftfrac *) malloc(sizeof(cftfrac)*nb);
        if(colgr->wgt     == NULL) {
            fprintf(Stderr, "*** cfcginit: memory allocation error\n");
            exit(1);
        }

        /* default values */
        for(i = 0; i < nb; i++)
            /* (K.K.) colgr->wgt[i]   = ((double) 0);*/
            cffrclr(&colgr->wgt[i]) ;

    } else {
        colgr->lvtx = NULL;

        colgr->tedge   = NULL;
        colgr->edge[0] = NULL;
        colgr->edge[1] = NULL;

        colgr->wgt     = NULL;
    }
}

/*==================================================================*/
Public  Void    cfcgfree(colgr)
cftgraph *colgr;
/*--------------------------------------------------------------------
 * Free allocated space in "colgr"
 */
{
    if(colgr->wgt != NULL) {
        free(colgr->wgt);
        colgr->wgt     = NULL;
    }

    if(colgr->edge[1] != NULL) {
        free(colgr->edge[1]);
        colgr->edge[1] = NULL;
    }

    if(colgr->edge[0] != NULL) {
        free(colgr->edge[0]);
        colgr->edge[0] = NULL;
    }

    if(colgr->tedge != NULL) {
        free(colgr->tedge);
        colgr->tedge   = NULL;
    }

    if(colgr->lvtx != NULL) {
        free(colgr->lvtx);
        colgr->lvtx    = NULL;
    }
}

/*==================================================================*/
Private int     cfvkind(nl, nk)
int     nl;
int     *nk;
/*--------------------------------------------------------------------
 * Classify a vertex which has "nl" legs.
 * The current classification should be considered as temporal one.
 * This function will be modified, especially for SUSY.
 */
{
    int i;
    int vtype = CFVDEL;
    int ok = True;

    if(nk[CFNPUNDEF] != 0) {
        fprintf(Stderr, "*** cfvkind : undefined particle\n");
        ok = False;
    }

    /* with quark and/or anti-quark */
    if(nk[CFQUARK] != 0 || nk[CFAQUARK] != 0) {

        if(nk[CFQUARK] != nk[CFAQUARK]) {
            fprintf(Stderr, "*** cfvkind : quark number ");
            fprintf(Stderr, "does not conserve\n");
            ok = False;

        /* We assume only one pair of quark and anti-quark in 
         * a vertex.
         */
        } else if(nk[CFQUARK] > 1) {
            fprintf(Stderr, "*** cfvkind : to many quarks\n");
            ok = False;

        /* Now there is a pair of quark and ant-quark */

        /* q + q-bar + ghost + ... */
        } else if(nk[CFGHOST] != 0 || nk[CFAGHOST] != 0) {
            fprintf(Stderr, "*** cfvkind : quark and ghost ");
            fprintf(Stderr, "appear in the same vertex\n");
            ok = False;

        /* q + q-bar + color-singlet-particles */
        } else if(nk[CFGLUON] == 0) {
            vtype = CFVQQ;

        /* q + q-bar + gluon */
        } else if(nk[CFGLUON] == 1) {
            if(nl == 3) {
                vtype = CFVQQG;
            } else {
                fprintf(Stderr, "*** cfvkind : quark, anti-quark, ");
                fprintf(Stderr, "gluon and others appear");
                fprintf(Stderr, "in the same vertex\n");
                ok = False;
            }

        /* q + q-bar + gluon + gluon + ... */
        } else {
            fprintf(Stderr, "*** cfvkind : illegal number ");
            fprintf(Stderr, "of gluon as qqgg...\n");
            ok = False;
        }

    /* with ghost/anti-ghost and without quarks */
    } else if(nk[CFGHOST] != 0 || nk[CFAGHOST] != 0) {

        /* we assume #ghost = #anti-ghost */
        if(nk[CFGHOST] != nk[CFAGHOST]) {
            fprintf(Stderr, "*** cfvkind : ghost number ");
            fprintf(Stderr, "does not conserve\n");
            ok = False;

        /* n(ghost=anti-ghost) + gluon + color-singlets */
        } else if(nk[CFGLUON] == 1) {

            /* ghost + anti-ghost + gluon */
            if(nl == 3) {
                vtype = CFVQQG;

            } else {
                fprintf(Stderr, "*** cfvkind : ghost, gluon");
                fprintf(Stderr, "and other particles\n");
                ok = False;
            }

        /* ghost + anti-ghost + color-singles */
        } else if(nk[CFGLUON] == 0) {

            /* counter term for ghost self-energy */
            if(nl == 2) {
                vtype = CFVHH;

            } else {
                fprintf(Stderr, "*** cfvkind : ghost, a-ghost");
                fprintf(Stderr, "and other particles\n");
                ok = False;
            }

        /* ghost + anti-ghost + n*gluon + color-singles, n > 1 */
        } else {
            fprintf(Stderr, "*** cfvkind : illegal number ");
            fprintf(Stderr, "of gluon as qqgg...\n");
            ok = False;
        }

    /* with gluons */
    } else if(nk[CFGLUON] != 0) {

        /* ggg */
        if(nk[CFGLUON] == 3 && nl == 3) {
            vtype = CFVGGG;

        /* gg */
        } else if(nk[CFGLUON] == 2 && nl == 2) {
            /* gluon counter term */
            vtype = CFVGG;

        } else {
            fprintf(Stderr, "*** cfvkind : illegal number ");
            fprintf(Stderr, "of gluon\n");
            fprintf(Stderr, "nl = %d\n", nl);
            for(i = 0; i < nl; i++) {
                fprintf(Stderr, "nk[%d] = %d\n", i, nk[i]);
            }
            ok = False;
        }

    /* only color-singlet particles */
    } else {
        vtype = CFVNONQCD;
    }

    if(!ok) {
        fprintf(Stderr, "*** cfvkind : illegal vertex \n");
        pvdump();
        prdump();
        exit(1);
    }
    return vtype;
}
