/*******************************************************************
 *    File : rdgrf.c
 *
 *    Read graph
 *
 *    Calling sequence:
 *
 *      1. Initialization
 *
 *              if(!rginit(<filename>))
 *                  exit(FAIL);
 *
 *      2. Read loop
 *
 *              while(rgproc(tproc*)) {
 *                  while(rgraph(tgraph*)) {
 *                    ...
 *                  }
 *              }
 *
 *      3. Close
 *
 *              rgclose();
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "tokenh.h"
#include "grdata.h"

/* Entry points */
Public  tgraph *rgalloc      ARG((tproc*, int));
Public  bool    rgclose      ARG((NoArg));
Public  Void    rgdump       ARG((tproc*, tgraph*));
Public  Void    rgigraph     ARG((tproc*, tgraph*));
Public  bool    rginit       ARG((string));
Public  bool    rgproc       ARG((tproc*));
Public  bool    rgraph       ARG((tgraph*));
Public  Void    rgsetv       ARG((tproc*,tgraph*,int,int,int*));

/* External references */
/* ergrf.c */
Extern  Void    ercont       ARG((NoArg));
Extern  Void    erdbg        ARG((string, string, int));
/* ilgrf.c */
Extern  int     ilmake       ARG((bool));
/* rdprbl.c */
Extern  pindex  pvfind       ARG((string));
Extern  Void    pvfname      ARG((string,string,int));
Extern  Void    pvread       ARG((string));
Extern  Void    pvreord      ARG((node));
Extern  int     pvgl4        ARG((node));
/* matchv.c */
Extern  vindex  pvmatc       ARG((int,int*));

/* Private functions */
Private bool    rgexternal   ARG((int));
Private bool    rgnode       ARG((node, int, int*));
Private pindex  rgene        ARG((int, string, int));
Private Void    rgndopt      ARG((int, int*, int*));
Private bool    rgenv        ARG((int,int,int,int*,string*,int*,int*));
Private Void    rgerror      ARG((string, string));

/* Local variable */
#define ELPART 0 
#define ELIORF 1
Local   char    rgmdl[MAXPATHLEN];
Local   int     eltbl[MAXELINE][ELIORF+1];
Local   int     edgetbl[MAXNODES][MAXNODES];
Local   FILE    *rgfp;
Local   int     nvertex;
Local   int     rgind;
Local   int     extcol;

/*==================================================================*/
Private Void    rgerror(fncn, msg)
string    fncn;
string    msg;
/*--------------------------------------------------------------------
 */
{
    fprintf(Stderr, "*err* %s : %s\n", fncn, msg);
}

/*==================================================================*/
Public  bool    rginit(fn)
string    fn;
/*--------------------------------------------------------------------
 *  initializes read package.
 *
 *   1. open grf file.
 *   2. read mdl file.
 */
{
    FILE*   echo;
    int     ind;
    trdsava sarea;
    static  int rdmdl=0;

#if DBG
    if(dbg >= rgdbg)
        printf("rginit(%s)\n", fn);
#endif

    /* echo : file pointer for echo of input. */
    echo = NULL;        /* no echo */

    /* open input file */
    rgfp = fopen(fn, "r");
    if(rgfp == NULL) {
        fprintf(Stderr, "*err* rginit : open err fname = %s\n", fn);
        return False;
    }

    /* initialize token package */
    rdinit(rgfp, RDNOCHANGE, echo);

    /* input : model="<model file>"; */
    rgind = rdstat();
    ind   = rdarg(rgind);
    fnmdl = rgmdl;
    if(rdchkid(rgind, "model") && rdtype(ind) == RDSTRING) {
        if(rdmdl == 0) {
            fnmdl = rgmdl;
            pvfname(fnmdl, rdid(ind), MAXPATHLEN);
            rdsave(&sarea);
            pvread(fnmdl);       /* read particle table */
            rdrestore(&sarea);   /* recover fp in the token package */
        } else {
            pvfname(fnmdl, rdid(ind), MAXPATHLEN);
            rgerror("rginit", "different model");
            fprintf(Stderr,"    previous : %s\n", fnmdl);
            fprintf(Stderr,"    new      : %s\n", rdid(ind));
            exit(1);
        }
        rdmdl++;
    } else {
        rgerror("rginit", "model = \"<filename>\"; is expected.");
        rdprint(rgind);
        ercont();
        return False;
    }

#if DBG
    if(dbg >= rgdbg) {
        printf("end of rginit\n");
    }
#endif
    return True;
}

/*==================================================================*/
Public  bool    rgproc(proc)
tproc *proc;
/*--------------------------------------------------------------------
 *    read header of a process
 *
 *    1. process number
 *    2. table of external particles
 *    3. order of coupling constants
 *    4. the number of loops
 *    5. flags.
 */
{
    int     ind, io, arg;
    bool    st;
    int     nextern, order, loop, onepi, expand, msize, gm2;
    string  kinem;

    cprocess = proc;
    nprocess  = -1;
    nextern   = -1;
    order     = 0;
    loop      = -1;
    onepi     = True;
    expand    = True;
    gm2       = True;
    kinem     = NULL;

    for(io = 0; io < nmordtbl; io++)
        cprocess->morder[io] = 0;

    st = True;
    while (True) {
        /* read : process=<int>; */
        rgind = rdstat();
        if(rdtype(rgind) != RDID)
            return False;

        ind = rdarg(rgind);

        /* read : process=<int>; */
        if(rdstrcmp(rdid(rgind), "end") == 0) {
            return False;
        } else if(rdstrcmp(rdid(rgind), "process") == 0) {
            if(rdtype(ind) == RDINT)
                nprocess = rdint(ind);
            else {
                rgerror("rginit", "process = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : external=<int>; */
        } else if(rdstrcmp(rdid(rgind), "external") == 0) {
            if(rdtype(ind) == RDINT)
                nextern = rdint(ind);
            else {
                rgerror("rginit", "external = <int>");
                rdprint(rgind);
                st = False;
            }
            /* read table of external particles */
            st = rgexternal(nextern);

        /* read : order=<int>; */
        /*
        } else if(rdstrcmp(rdid(rgind), "order") == 0) {
            if(rdtype(ind) == RDINT)
                order = rdint(ind);
            else {
                rgerror("rginit", "order = <int>");
                fprintf(Stderr,"*err* rginit : order = <int>\n");
                rdprint(rgind);
                st = False;
            }
        */

        /* read : loop=<int>; */
        } else if(rdstrcmp(rdid(rgind), "loop") == 0) {
            if(rdtype(ind) == RDINT)
                loop = rdint(ind);
            else {
                rgerror("rginit", "loop = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : opi=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "opi") == 0) {
            if(!rdbool(ind, &onepi)) {
                rgerror("rginit", "opi = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : expand=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "expand") == 0) {
            if(!rdbool(ind, &expand)) {
                rgerror("rginit", "expand = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : gm2=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "gm2") == 0) {
            if(!rdbool(ind, &gm2)) {
                rgerror("rginit", "gm2 = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : kinem=string; */
        } else if(rdstrcmp(rdid(rgind), "kinem") == 0) {
            arg = rdarg(rgind);
            if(rdtype(arg) == RDSTRING){
                kinem = (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(kinem == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(kinem, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rgind);
                st = False;
            }

        } else {
            for(io = 0; io < nmordtbl; io++) {
                if(rdstrcmp(rdid(rgind), mordtbl[io]) == 0) {
                    cprocess->morder[io] = rdint(ind);
                    order += cprocess->morder[io];
                    break;
                }
            }
            if(io >= nmordtbl) {
                break;
            }
        }
    }
    if (nprocess < 0) {
        rgerror("rginit", "process = <int>");
        st = False;
    }
    if (nextern < 0) {
        rgerror("rginit", "external = <int>");
        st = False;
    }
    if (order <= 0) {
        rgerror("rginit", "order = <int>");
        fprintf(Stderr,"*err* rginit : \"order\"=<int>\n");
        st = False;
    }
    if (loop < 0) {
        rgerror("rginit", "loop = <int>");
    }
 
    /* save the process */
#if DBG
    if(dbg >= rgdbg)
        printf("st = %d\n", st);
#endif
    if(st) {
        if(proc == NULL) {
            rgerror("rgproc", "null proc pointer");
            exit(1);
        }
        cprocess->procid    = nprocess;
        cprocess->graphs    = 0;
        cprocess->nextern   = nextern;
        cprocess->oextern   = nextern;
        cprocess->order     = order;
        cprocess->loop      = loop;
        cprocess->onepi     = onepi;
        cprocess->expand    = expand;
        cprocess->gm2       = gm2;
        cprocess->kinem     = kinem;
        cprocess->maxnodes  = nextern +   order;
        cprocess->maxlegs   = nextern + 3*order;
        msize =   sizeof(tgraph) 
                + sizeof(tnode) * cprocess->maxnodes 
                + sizeof(tleg)  * cprocess->maxlegs;
        cprocess->msize     = msize;
        cprocess->extcol    = extcol;   /* see rgexternal() */

        /* order of QCD coupling constant */
        cprocess->qcd = 0;
        for(io = 0; io < nmordtbl; io++) {
            if(rdstrcmp("qcd", mordtbl[io]) == 0) {
                cprocess->qcd = cprocess->morder[io];
                break;
            }
        }
    }


#if DBG
    if(dbg >= rgdbg) {
        printf("end of rgproc\n");
    }
#endif
    return st;
}

/*==================================================================*/
Public  bool    rgclose(NoArg)
/*--------------------------------------------------------------------
 *  closes read packages.
 */
{
#if DBG
    if(dbg >= rgdbg)
        printf("rgclose()\n");
#endif
    if(rgfp != NULL) {
        rdend();
        fclose(rgfp);
        rgfp = NULL;
    }

    return True;
}

/*==================================================================*/
Private bool    rgexternal(nextern)
int nextern;
/*--------------------------------------------------------------------
 *  read the table of external particles.
 */
{
    int i, a, iorf, n;
    bool st;
    pindex pc;

#if DBG
    erdbg("rgexternal", "begin", rgdbg);
    if(dbg >= rgdbg) {
        printf("rgexternal(%d)\n", nextern);
    }
#endif

    st = True;
    extcol = 0;
    for(i = 0; i < nextern; i++) {
        rgind = rdstat();

        /* external particle number */
        n = tonode(rdint(rgind));
        if (i != n) {
            fprintf(Stderr, "*err* rgexternal : illegal order\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* initial / final */
        a = rdarg(rgind);
        if(rdchkid(a, "initial"))
            iorf = ELINITIAL;
        else if(rdchkid(a, "final"))
            iorf = ELFINAL;
        else {
            fprintf(Stderr, "*err* rgexternal : ");
            fprintf(Stderr, "not initial nore final\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* particle name */
        a = rdarg(a);
        if(rdtype(a) != RDID) {
            fprintf(Stderr, "*err* rgexternal : ");
            fprintf(Stderr, "illegal particle name.\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* save */
        pc = rgene(n, rdid(a), iorf);    /* when error stop */
        if(ptable[abs(pc)]->color != 1)
            extcol++;
#if DBG
        if(dbg >= rgdbg)
            rdprint(rgind);
#endif
    }

    if(!st) {
        while (!rdchkid(rgind, "eend"))
            rgind = rdstat();
        exit(1);
    }
    rgind = rdstat();
    st = rdchkid(rgind, "eend");

#if DBG
    erdbg("rgexternal", "end", rgdbg);
#endif

    return st;
}

/*==================================================================*/
Public  bool    rgraph(grf)
tgraph  *grf;
/*--------------------------------------------------------------------
 *  returns False if EOF or error is found.
 */
{
    int i, j, lp, lg;
    bool st;
    int  legs;
    long grn;
    int     nlgtbl;  /* counter of allocated legs */
    node nd;

#if DBG
    erdbg("rgraph", "begin", rgdbg);
#endif
    st = True;

    if(grf == NULL) {
        rgerror("rgraph", "null graph\n");
        exit(1);
    }

    rgigraph(cprocess, grf);

    nlgtbl = 0;  /* the number of allocated legs */

    cgraph = grf;
    ndnodes() = cprocess->nextern;  /* necessary ??? */

    /* "graph=<int>;" */
    if(rdkval(rgind, "pend", &grn)) {
        return False;
    } else if(!rdkval(rgind, "graph", &grn)) {
        fprintf(Stderr, "*err* rgraph : graph=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->grid  = grn;
    cgraph->subid = 0;

    /* "gtype=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "gtype", &grn)) {
        fprintf(Stderr, "*err* rgraph : gtype=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->gtype = grn;

    /* "sfactor=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "sfactor", &grn)) {
        fprintf(Stderr, "*err* rgraph : sfactor=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->sfactor = grn;

    /* "vertex=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "vertex", &grn)) {
        fprintf(Stderr, "*err* rgraph : vertex=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    nvertex = grn;

    ndnodes() = cgraph->conjsrt = cprocess->nextern + nvertex;

    /* read nodes line */
    for(i = 0; i < ndnodes(); i++) {
        rgind = rdstat();
        if(!rgnode(i, ndnodes(), &nlgtbl)) {
#if DBG
            if(dbg >= rgdbg)
                printf("st of rgnode = False\n");
#endif
            st = False;
            break;
        }
#if DBG
        if(dbg >= rgdbg) {
            printf("node = %d ",i);
            rdprint(rgind);
        }
#endif
    }

    /* "vend;" */
    rgind = rdstat();
    if(!rdchkid(rgind, "vend")) {
        fprintf(Stderr, "*err* rgraph : vend;");
        fprintf(Stderr, " nnodes = %d, nexternal = %d, nvertex = %d\n",
                ndnodes(), cprocess->nextern, nvertex);
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        return False;
    }

    /* "gend=<number>;" */
    rgind = rdstat();
    if(!rdchkid(rgind, "gend")) {
        fprintf(Stderr, "*err* rgraph : gend;\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        return False;
    }

    rgind = rdstat();

    /* construct internal line table and ndnextn(*,*) */
    cgraph->nilines = ilmake(True);

    /* construct edge table */
    for(i = 0; i < ndnodes(); i++)
        for(j = 0; j < ndnodes(); j++)
            edgetbl[i][j] = 0;

    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            edgetbl[nd][ndnextn(nd,lg)]++;
        }
    }

    /* calculate the number of loops */
    legs = 0;
    for(i = 0; i < ndnodes(); i++) {
        for(j = 0; j < ndnodes(); j++) {
            legs += edgetbl[i][j];
        }
    }

    lp = cprocess->loop;
    for(i = 0; i < ndnodes(); i++)
        lp -= ndloop(i);

    if(lp != legs/2 - ndnodes() + 1) {
        fprintf(Stderr,"*err* rgraph : inconsistent loop\n");
        fprintf(Stderr,"lp = %d, legs = %d, nnodes = %d\n",
               lp, legs, ndnodes());
        exit(1);
    }

    if(st)
        cprocess->graphs++;

#if DBG
    erdbg("rgraph", "end", rgdbg);
#endif

    return st;
}

/*==================================================================*/
Public  Void    rgigraph(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 *  initialize graph data
 */
{
    int  ptr;

    ptr        = sizeof(tgraph);
    grf->basen = ptr;
    ptr       += sizeof(tnode)*(proc->maxnodes);
    grf->offsl = ptr;

    grf->nnodes = 0;
    grf->ngl4   = 0;

#if DBG
    if(dbg>=rgdbg)
        rgdump(cprocess, cgraph);
#endif
}

/*==================================================================*/
Public  tgraph *rgalloc(proc, ngrf)
tproc *proc;
int       ngrf;
/*--------------------------------------------------------------------
 *  Allocate area for "ngrf" graphs to "grf".
 */
{
    int     n, msize;
    char   *ptr;
    tgraph *grf;

    msize = proc->msize;

    if((ptr = (char *) malloc(ngrf*msize)) == NULL) {
        fprintf(Stderr, "rgalloc : memory allocation failed.\n");
        exit(16);
    }
    grf = (tgraph *) ptr;
    for(n = 0; n < ngrf; n++) {
#if DBG
        if(dbg >= rgdbg)
            printf("rgalloc : ptr[n] = %p\n", n, ptr);
#endif
        rgigraph(proc, (tgraph *) ptr);
        ptr += msize;
    }

    return grf;
}
/*==================================================================*/
Private bool    rgnode(nd, nnd, nlgtbl)
node    nd;
int     nnd;
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  reads a node and its connections.
 */
{
    int    legn[MAXNODES];
    string legp[MAXNODES];
    int    mord[MAXMORD];
    int  nleg, a, b, lp, l;
    bool st;

#if DBG
    erdbg("rgnode", "begin", rgdbg);
    if(dbg >= rgdbg)
        printf("rgnode(%d, %d)\n", nd, nnd);
#endif
    /* node number */
    if(rdtype(rgind) != RDINT) {
        rgerror("rgnode", "number is expected");
        rdprint(rgind);
        rdpstat(rgind, 0);
        return False;
    }
    if (nd != tonode(rdint(rgind))) {
        rgerror("rgnode", "illegal node number.");
        rdprint(rgind);
        rdpstat(rgind, 0);
        return False;
    }

    /* options to the node */
    lp = 0;
    if((a = rdopt(rgind)) >= 0)
        rgndopt(a, &lp, mord);

    /* legs */
    nleg = 0;
    for (a = rdarg(rgind); a >= 0; a = rdarg(a)) {
        if(rdtype(a) != RDINT) {
            rgerror("rgnode", "not i-line number for a leg.");
            rdprint(rgind);
            return False;
        }
        if((legn[nleg] = rdint(a)) < 1) {
            rgerror("rgnode", "illegal i-line number for a leg.");
            rdprint(rgind);
            return False;
        }

        b = rdopt(a);

        if(rdtype(b) == RDID) {
            legp[nleg++] = rdid(b);
        } else {
            legp[nleg++] = "undef";
        }
    }

#if DBG
    if(dbg >= rgdbg) {
        printf("nd = %d, nleg = %d\n", nd, nleg);
        for(l = 0; l < nleg; l++)
            printf(" %d", legn[l]);
        printf("\n");
    }
#endif

     /* save */
    st = rgenv(nd, nleg, lp, legn, legp, mord, nlgtbl);  /* save */

#if DBG
    if(dbg >= rgdbg) {
        printf("rgnode : st = %d\n", st);
        rdprint(rgind);
    }
#endif

#if DBG
    erdbg("rgnode", "end", rgdbg);
#endif

    return st;
 }

/*==================================================================*/
Private pindex  rgene(n, ptcl, iorf)
int n;
string ptcl;
int iorf;
/*--------------------------------------------------------------------
 *  save external particle.
 */
{
    pindex pc;

#if DBG
    if(dbg >= rgdbg)
        printf("rgene(%d, %s, %d)\n", n, ptcl, iorf);
#endif
    if(n >= MAXELINE) {
        rgerror("rgene", "too many e-lines");
        exit(1);
    }

    pc = pvfind(ptcl);
    if(iorf == ELINITIAL)
        pc *= ptable[abs(pc)]->antip;


    eltbl[n][ELIORF] = iorf;
    eltbl[n][ELPART] = pc;

    return pc;
}

/*==================================================================*/
Public  Void    rgsetv(proc, grf, v, lg, nlgtbl)
tproc  *proc;
tgraph *grf;
int    v;
int    lg;
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  setup table of legs.
 */
{
    if(*nlgtbl + lg > proc->maxlegs) {
        rgerror("rgsetv", "leg table overflow");
        fprintf(Stderr, " %d+%d > %d\n", *nlgtbl, lg, proc->maxlegs);
        rgdump(proc, NULL);
        exit(1);
    }

    nebasel(grf, v) = grf->offsl + (*nlgtbl)*sizeof(tleg);
    *nlgtbl += lg;
}

/*==================================================================*/
Private bool    rgenv(v, lg, lp, legn, legp, mord, nlgtbl)
int    v;
int    lg;
int    lp;
int    legn[];
string legp[];
int    mord[];
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  save node
 */
{
    node n;
    int  p;
    bool pdef;

#if DBG
    if(dbg >= rgdbg) {
        printf("rgenv(%d, %d, %d, %p)\n", v, lg, lp, legn);
        printf("rgenv : %d+%d > %d\n", *nlgtbl, lg, cprocess->maxlegs);
    }
#endif

    rgsetv(cprocess, cgraph, v, lg, nlgtbl);

    if(v < cprocess->nextern) {
        ndtype(v)   = NDEXTERN;
        ndlegs(v)   = 1;
        ndloop(v)   = 0;
        ndiorf(v)   = eltbl[v][ELIORF];
        if(lg != 1) {
            rgerror("rgenv", "inconsistent e-line");
            exit(1);
        }
    } else {
        ndtype(v)   = (lp == 0) ? NDTREEV :NDLOOPEDV;
        ndlegs(v)   = lg;
        ndloop(v)   = lp;
        ndvertex(v) = VNULL;
        ndgl4(v)    = 0;
    }

    for(n = 0; n < ndnodes(); n++)
        edgetbl[v][n] = 0;
    pdef = True;
    for(p = 0; p < lg; p++) {
        ndiline(v, p)    = legn[p];
        ndparticle(v, p) = pvfind(legp[p]);
        if(ndparticle(v, p) == PNULL)
            pdef = False;
    }
    if(v >= cprocess->nextern) {
        if(pdef) {
            if((ndvertex(v) = pvmatc(v, mord)) != VNULL)
                pvreord(v);
            if((ndgl4(v) = pvgl4(v)) == 4)
                ndngl4()++;
        } else {
            ndvertex(v) = VNULL;
            ndgl4(v) = 0;
        }

/*      not exclude unedfined vertex type

        if(ndvertex(v) == VNULL) {
            rgerror("rgenv", "vertex is not defined in the model.");
            rdprint(rgind);
            return False;
        }
*/
    }

    return True;
}

/*==================================================================*/
Private Void rgndopt(opt, lp, mord)
int opt;
int *lp;
int *mord;
/*--------------------------------------------------------------------
 *  read option for the node
 *
 *    *lp   : loop  <= "loop=<int>;"
 *    *mord : order <= "order={<int>,...};"
 */
{
    int  arg, i;

#if DBG
    if(dbg >= rgdbg)
        printf("rgndopt(%d, %p)\n", opt, lp);
#endif

    *lp = 0;
    for ( ; opt >= 0; opt = rdbrothr(opt)) {
        if(rdtype(opt) == RDID) {
            if(rdstrcmp(rdid(opt),"loop") == 0) {
                arg = rdarg(opt);
                if(rdtype(arg) != RDINT) {
                    fprintf(Stderr, "*err* rgndopt: illegal loop\n");
                } else {
                    *lp = (int)rdint(arg);
                }
            } else if(rdstrcmp(rdid(opt), "order") == 0) {
                /* do not redefine */
                arg = rdarg(opt);
                for (i = 0; i < nmordtbl; i++) {
                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "*err* rgndopt: illegal order\n");
                    } else {
                        mord[i] = (int)rdint(arg);
                    }
                    if(rdtype(arg) != RDEOF)
                        arg = rdarg(arg);
                }
            }
        }
    }
}

/*==================================================================*/
Public  Void    rgdump(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 *  print the current graph
 */
{
    int n, p, io;

    printf("**** rgdump ****\n");

    printf("process : *(%p)", proc);
    if(proc == NULL) {
        printf(" NULL process\n");
        return;
    }
    printf(": id = %d\n",     proc->procid  );
    printf("  graphs = %d\n", proc->graphs  );
    printf("  nextern = %d",  proc->nextern );
    printf("  loop = %d",     proc->loop    );
    printf("  order = %d\n",    proc->order   );
    for(io = 0; io < nmordtbl; io++)
        printf("  %s = %d,", mordtbl[io], proc->morder[io]);
    putchar(EOL);
    printf("  expand = %d",   proc->expand  );
    printf("  onepi = %d",  proc->onepi   );
    printf("  gm2 = %d\n",   proc->gm2  );
    printf("  kinem = %s\n",  proc->kinem   );
    printf("  maxnodes = %d", proc->maxnodes);
    printf("  maxlegs = %d",  proc->maxlegs );
    printf("  msize = %d\n",  proc->msize   );
    printf("  qcd = %d",      proc->qcd     );
    printf("  extcol = %d\n", proc->extcol  );
  

    printf("graph : *(%p)", grf);
    if(grf == NULL) {
        printf(" NULL graph\n");
        return;
    }

    cgraph   = grf;

    if(grf->subid == 0)
        printf(": id=%d\n",      grf->grid);
    else
        printf(": id=%d(%d)\n",  grf->grid, grf->subid);
    printf("  nnodes = %d",  grf->nnodes);
    printf("  ngl4   = %d",  grf->ngl4);
    printf("  basen = %d",   grf->basen);
    printf("  offsl = %p\n", grf->offsl);

    for(n = 0; n < grf->nnodes; n++) {
        printf("  node = %d *(%p)", n, ndnode(0));
        printf("  type = %d", ndtype(n));
        printf("  loop = %d",       ndloop(n));
        printf("  kind = %d",     ndvertex(n));
        printf("  gl4  = %d\n",     ndgl4(n));
        printf("  nlegs = %d",      ndlegs(n));
        printf("  legs = *(%p)\n",  ndleg(n, 0));
        if(ndleg(n, 0) == NULL)
            break;
        printf("  %d:", n);
        for(p = 0; p < ndlegs(n); p++)
            printf(" %d[%d:%d]", ndiline(n, p), ndnextn(n, p), 
                   ndparticle(n, p));
        putchar(EOL);
    }

    printf("**** end of rgdump ****\n");
}

